"""This module stores base classes and functions.

Attributes:
    RESFILEDS (dict): stores results fields under results type

        key (str):
            - 'displ': displacement
            - 'stress': stress
            - 'strain': strain
            - 'force': force
        value (list):
            - 'displ' = 'ux','uy','uz','utot'
            - 'stress' = 'Sx','Sy','Sz','Sxy','Syz','Szx','Seqv','S1','S2','S3'
            - 'strain' = 'ex','ey','ez','exy','eyz','ezx','eeqv','e1','e2','e3'
            - 'force' = 'fx','fy','fz'
    FIELDTYPE (dict): the inverse dict of RESFIELDS

        For example: key 'ux' --> value: 'displ'
"""
from math import ceil
from . import environment

RESFIELDS = {}
RESFIELDS['displ'] = 'ux,uy,uz,utot'.split(',')
RESFIELDS['stress'] = 'Sx,Sy,Sz,Sxy,Syz,Szx,Seqv,S1,S2,S3'.split(',')
RESFIELDS['strain'] = 'ex,ey,ez,exy,eyz,ezx,eeqv,e1,e2,e3'.split(',')
RESFIELDS['force'] = 'fx,fy,fz'.split(',')

#FIELDTYPE is a dict that inverts the RESFIELDS dictionary mapping.
FIELDTYPE = {}
for (k, v) in RESFIELDS.items():
    for vi in v:
        FIELDTYPE[vi] = k

class Idobj(object):
    """Makes an object that stores an id number.

    This is a base class for nodes, lines, areas etc.

    Attributes:
      id (int): the unique id number for the item
    """

    def __init__(self):
        self.id = -1

    def set_id(self, id):
        """Sets the id number """
        self.id = id

    def __hash__(self):
        """Sets the hash of the item to the id number.

        This allows one to make sets of these items.
        """
        return self.id

def plot_finish(plt, fname, display):
    """Display and or save plot."""
    if fname != '':
        # save the image
        fname += '.png'
        if environment.DPI != None:
            plt.savefig(fname, dpi=environment.DPI, bbox_inches='tight')
        else:
            plt.savefig(fname, bbox_inches='tight')
        print('File %s was saved.' % fname)

    if display:
        plt.tight_layout()
        plt.show()

    # remove all figures
    plt.close()

def plot_set_bounds(plt, axials, radials):
    """Sets the axial and radial bounds of the shown plot."""
    vert = max(radials) - min(radials)
    horiz = max(axials) - min(axials)
    vadder = (vert)/5
    hadder = (horiz)/5
    (vmax, vmin) = (max(radials)+vadder, min(radials)-vadder)
    (hmax, hmin) = (max(axials)+hadder, min(axials)-hadder)
    plt.xlim(hmin, hmax)
    plt.ylim(vmin, vmax)

class Itemlist(list):
    """Makes a custom list used to store lists of non-mesh items.

    non-mesh items = Point, Line, Arc, Area, Part, Components, Loads
    This class allows us to automatically assign id numbers to items in the
    list. Thes id numbers are needed when we send geometry out for meshing.
    All instances of this class start as empty lists.
    """

    def __init__(self):
        super().__init__() # these lists start empty

    def get_ids(self):
        """Returns a list of ids. Loops through all items in self."""
        return [a.id for a in self]

    def get_next_id(self):
        """Returns the next unused id."""
        ids = self.get_ids()
        minid = 0
        if len(ids) == 0:
            return minid    # list is empty so return the minid number
        else:
            ids = sorted(ids)
            maxid = ids[-1]
            unused = list(set(list(range(minid, maxid+2))) - set(ids))
            return unused[0]    # return first available id number

    def append(self, item):
        """Adds an item to the list and sets the item's id.

        Args:
          item (Point or Line or Arc or Area or Part): item to add to the list
        """
        idnum = self.get_next_id()
        item.set_id(idnum)
        super().append(item)
        return item


class Meshlist(list):
    """Makes a custom list used to store lists of nodes and elements.

    All instances of this class start as empty lists.
    """

    def __init__(self):
        list.__init__([])

    def get_minid(self):
        """Returns the min id number in the list."""
        ids = [x.id for x in self]
        return min(ids)

    def get_maxid(self):
        """Returns the max id number in the list."""
        ids = [x.id for x in self]
        return max(ids)

    def idget(self, idnum):
        """Returns an item with the passed id number.

        Args:
          idnum (int): passed id number, we want the item that has this

        Returns:
          item or None: returns item if found, None if not found
        """
        for item in self:
            if item.id == idnum:
                return item
        #print ('ID %i was not found!' % (idnum))
        return None

    def set_minid(self, val):
        """Sets the min id to the passed val.

        Args:
          val (int): New min id number in the list
        """
        print('Re-indexing elements to min element number of: %i' % val)
        print('Old min:%i  max:%i' % (self.get_minid(), self.get_maxid()))
        minid = self.get_minid()
        offset = minid - val
        for item in self:
            item.id -= offset
        print('New min:%i  max:%i' % (self.get_minid(), self.get_maxid()))


def chunk_list(inlist, size):
    """Returns a list of lists where each list <= size length.

    Splits inlist into list of lists where each child list len <= size.

    Args:
      inlist (list): list that we want cut into smaller lists
      size (int): max length of small lists returned

    Returns:
      res (list of lists): list of list where each child list <= size length
    """
    res = []
    numlists = ceil(len(inlist)/size)
    for ind in range(numlists):
        res.append(inlist[ind*size:(ind+1)*size])
    return res

def listify(items):
    """Returns a list of items. If items is an object it puts it in a list.

    If a list is passed, it returns the list without changing it.
    If an object is passed, it returns a list with the object in it.

    Args:
        items (object or list): item or list to convert to a list
    """
    if not isinstance(items, list):
        items = [items]
    return items



"""This module stores load classes.
"""
from math import pi

class ConstLoad(object):
    """Makes a load. Many types are supported.

    Possible load types:
        - forces
        - displacements
        - thickness (on nodes)
        - matl (on elements)
        - pressure (or stress on faces)
        - gravity (on elements)
        - rotation about an axis (on elements)

    Args:
        ltype (string): load type:

            - 'fx','fy','fz': force on each axis
            - 'ux','uy','uz': displacement on each axis
            - 'nodal_thickness': thickness on element nodes
            - 'matl': matl on elements
            - 'press': pressure, + is tension, - is compresion
            - 'gravity': gravity in x axis direction, - goes towards y axis
            - 'radps': radians per second rotation
            - 'rpm': rotations per minute rotation

        comp (Component): component that the load is applied to
        val (double or Matl): value of the load
            Matl is needed when setting material loads
            otherwise a double value is used to describe the load
    """

    def __init__(self, ltype, comp, val):
        self.ltype = ltype
        self.comp = comp
        self.val = val

    def get_list(self):
        """Returns a list of [item, val] for plotting."""
        res = []
        if self.ltype == 'press':
            # pressure
            faces = self.comp.get_children()
            for face in faces:
                res.append([face, self.val])
        elif self.ltype in ['ux', 'uy', 'uz']:
            nodes = self.comp.get_children()
            load_dict = {self.ltype:self.val}
            for node in nodes:
                res.append([node, load_dict])
        return res

    def ccx(self):
        """Writes a load for Calculix ccx solver.

        Returns a list strings, where each item is a line to write
        to a Calculix .inp text file.
        """
        res = []
        cname = self.comp.name
        if self.ltype == 'press':
            # pressure
            res.append('** Pressure on component: '+cname)
            res.append('*DLOAD')
            faces = self.comp.get_children()
            for face in faces:
                myline = '%i,P%i,%f' % (face.element.id, face.id, self.val)
                res.append(myline)
        elif self.ltype == 'gravity':
            # gravity
            res.append('** Gravity on component: '+cname)
            res.append('*DLOAD')
            res.append('%s,GRAV,%f,-1.,0.,0.' % (cname, self.val))
        elif self.ltype[0] == 'u':
            # displacement
            axis = self.ltype[1]
            anum = {'x':1, 'y':2, 'z':3}
            axis = anum[axis]
            res.append('*BOUNDARY')
            res.append('%s,%i,%i,%f' % (cname, axis, axis, self.val))
        elif self.ltype[0] == 'f':
            # force
            res.append('** Force on component: '+cname)
            nodes = len(self.comp.items)
            fval = self.val/nodes
            axis = self.ltype[1]
            anum = {'x':1, 'y':2, 'z':3}
            axis = anum[axis]
            res.append('*CLOAD')
            res.append('%s,%i,%f' % (cname, axis, fval))
        elif self.ltype == 'nodal_thickness':
            res.append('*NODAL THICKNESS')
            res.append('%s, %f' % (cname, self.val))
        elif self.ltype == 'radps':
            radsq = self.val**2 # square the speed in radians per sec
            res.append('** Rotation force on component: '+cname)
            res.append('** Speed = '+str(self.val)+' radians/sec')
            res.append('*DLOAD')
            res.append('%s,CENTRIF,%f,0.,0.,0.,0.,1.,0.' % (cname, radsq))
        elif self.ltype == 'rpm':
            rad = 2.0*pi*(self.val/60.0)
            radsq = rad**2 # square the speed in radians per sec
            res.append('** Rotation force on component: '+cname)
            res.append('** Speed = '+str(self.val)+' rotations/minute')
            res.append('*DLOAD')
            res.append('%s,CENTRIF,%f,0.,0.,0.,0.,1.,0.' % (cname, radsq))
        elif self.ltype == 'matl':
            mat = self.val.name
            res.append('*SOLID SECTION,ELSET=%s,MATERIAL=%s' %
                       (cname, mat))

        return res


class LinearLoad(object):
    """Makes a load which varies depending on x, y, or z location.

    Load Equation:
        Load(x) = const + mult*(xo - x)

    This load is used to set water pressure, where p = po + rho*g*(xo - x)
    For water pressure this becomes:

        - P(x) = po + (rho*g)*(xo-x)
        - P(x) = po + mult*(xo-x)
        - g is positive, xo > x

    Args:
        ltype (str): load type 'press_fluid' is the only valid option
        comp (Component): component that the load acts on
        const (float): constant term in load equation
        mult (float): mult term in the load equation
        xo (float): xo term in the load equation
        axis (str): axis that the load depends on
            'x' or 'y' are valid options
    """

    def __init__(self, ltype, comp, const, mult, xo, axis='x'):
        self.ltype = ltype
        self.comp = comp
        self.const = const
        self.mult = mult
        self.xo = xo
        self.axis = axis

    def get_val(self, xval):
        """Returns the load function value at the given location.

        Args:
            xval (float): the location where we want to find the load
        """
        res = self.const + (self.xo - xval)*self.mult
        return res

    def get_list(self):
        """Returns a list of [item, val] for plotting."""
        res = []
        faces = self.comp.get_children()
        for face in faces:
            xvals = [getattr(n, self.axis) for n in face.nodes]
            xavg = sum(xvals)/len(xvals)
            pval = self.get_val(xavg)
            res.append([face, pval])
        return res

    def ccx(self):
        """Writes a load for Calculix ccx solver.

        Returns a list strings, where each item is a line to write
        to a Calculix .inp text file.
        """
        res = []
        res.append('** Fluid pressure on component: '+self.comp.name)
        faces = self.comp.get_children()
        for face in faces:
            res.append('*DLOAD')
            xvals = [getattr(n, self.axis) for n in face.nodes]
            xavg = sum(xvals)/len(xvals)
            pval = self.get_val(xavg)
            myline = '%i,P%i,%f' % (face.element.id, face.id, pval)
            res.append(myline)
        return res



#!/usr/bin/python
"""
pycalculix Description:
This module provides a python interface to the Calculix FEA preprocessor and
solver. Simplicity is emphasized.

2D models are supported including:
plane stress, plane strain, and axisymmetric problems

Library Capabilities:
Build Geometry
Mesh Geometry
Apply Loads
Solve Model
View Results
Query Results
Select Sub-Portion of Model and view those results

See documentation and examples at:
http://justinablack.com/pycalculix/
https://github.com/spacether/pycalculix
"""

__author__ = "Justin Black"
__copyright__ = "Copyright 2015, Justin Black"
__credits__ = ["Justin Black"]
__license__ = "Apache 2.0"
__maintainer__ = "Justin Black"
__email__ = "justin.a.black[insert~at~sign]gmail.com"
__status__ = "Beta"
__all__ = ['environment', 'geometry', 'components', 'loads', 'partmodule',
           'mesh', 'material', 'feamodel', 'problem', 'results_file',
           'cadimporter']
from .version import __version__

from .feamodel import FeaModel
from .partmodule import Part
from .material import Material
from .problem import Problem
from .results_file import ResultsFile
from .cadimporter import CadImporter



"""This module stores geometry classes, which are used to make parts.
"""
from math import atan2, pi, cos, sin, radians, degrees
from matplotlib.patches import Arc as AArc
from matplotlib.patches import PathPatch
from matplotlib.path import Path
from numpy.linalg import det
from numpy import array
from numpy.linalg import solve as linsolve

from . import base_classes

#accuracy for small numbers in math below
ACC = .00001 # original

LDIVS = {}

def get_text_hv(angle):
    """Returns (ha, va) text horizontal and vertical alignment for line label.

    This makes the line label text inside its area, assuming areas closed CW.

    Args:
        angle (float): line normal vector in degrees, between 180 and -180
    """
    horiz, vert = ('', '')
    if abs(angle) < 67.5:
        horiz = 'right'
    elif 67.5 <= abs(angle) < 112.5:
        horiz = 'center'
    elif 112.5 <= abs(angle) <= 180:
        horiz = 'left'
    if abs(angle) < 22.5 or abs(angle) > 157.5:
        vert = 'center'
    else:
        if angle > 0:
            vert = 'top'
        else:
            vert = 'bottom'
    return (horiz, vert)


class Point(base_classes.Idobj):
    """Makes a point or vector.

    Args:
        x (float): vertical coordinate of point
        y (float): horizontal coordinate of point
        z (float): in-page coordinate of point, defaults to zero

    Attributes:
        x (float): vertical coordinate of point
        y (float): horizontal coordinate of point
        z (float): in-page coordinate of point, defaults to zero
            nodes (Node): a list of nodes in a mesh at this point.
            List length is 1.
        nodes (list): list of nodes under point
        lines (set): set of lines which use this point
        arc_center (bool): True if this point is an arc center
    """

    def __init__(self, x, y, z=0):
        self.x = x
        self.y = y
        self.z = z
        self.nodes = []
        self.lines = set()
        self.arc_center = False
        self.on_part = True
        self.esize = None
        base_classes.Idobj.__init__(self)

    def get_name(self):
        """Returns item name."""
        return 'P'+str(self.id)

    def __eq__(self, other):
        """Checks x and y equality to other point.

        Args:
            other (Point): passed point
        """
        if isinstance(other, Point):
            if self.x == other.x and self.y == other.y:
                return True
            return False
        else:
            return False

    def __add__(self, other):
        """Returns new point = self + other.

        Args:
            other (Point): passed point

        Returns:
            Point = self + other (new point, does not modify self)
        """
        return Point(self.x+other.x, self.y+other.y, self.z+other.z)

    def __sub__(self, other):
        """Returns new point = self - other.

        Args:
            other (Point): passed point

        Returns:
            Point = self - other (new point, does not modify self)
        """
        return Point(self.x-other.x, self.y-other.y, self.z-other.z)

    def __hash__(self):
        """Set the hash of the item to the id number.

        This allows one to make sets of these items.
        """
        return self.id

    def __mul__(self, factor):
        """Returns new point = self*factor.

        Args:
            factor (float): factor to multiply the Point by

        Returns:
            Point = self*factor (new point, does not modify self)
        """
        return Point(self.x*factor, self.y*factor, self.z*factor)

    def __truediv__(self, other):
        """Returns new point = self/other.

        Args:
            other (float or Point): item we're dividing by

        Returns:
            Point or factor. Does not modify self.
        """
        if isinstance(other, float):
            return Point(self.x/other, self.y/other, self.z/other)
        elif isinstance(other, Point):
            res = []
            if other.x != 0:
                res.append(self.x/other.x)
            if other.y != 0:
                res.append(self.y/other.y)
            if other.z != 0:
                res.append(self.z/other.z)
            factor = sum(res)/len(res)
            return factor

    def set_line(self, line):
        """Saves the line or arc in the point."""
        if line not in self.lines:
            self.lines.add(line)
        if self.arc_center == True:
            if len(self.lines) > 1:
                for line in self.lines:
                    if isinstance(line, Line):
                        self.arc_center = False
                        break

    def unset_line(self, line):
        """Removes the line or arc in the point."""
        if line not in self.lines:
            return
        self.lines.remove(line)

    def length(self):
        """Returns the length of this point or vector."""
        res = (self.x**2 + self.y**2)**0.5
        return res

    def label(self, axis):
        """Labels the point on a Matplotlib axis.

        Args:
            axis (Matplotlib Axis): Matplotlib Axis
        """
        axis.annotate(self.get_name(), (self.y, self.x))

    def plot(self, axis, label=True):
        """Plots the point on the passed matplotlib axis.

        Args:
            axis (Matplotlib axis): plate to plot the point
            pnum (bool): True turns on point labeling
        """
        axis.scatter(self.y, self.x)
        if label:
            self.label(axis)

    def make_unit(self):
        """Modifies self vector to make it a unit vector."""
        lenval = self.length()
        self.x = self.x*1/lenval
        self.y = self.y*1/lenval

    def ang_rad(self):
        """Returns angle in radians, assume point is vector from 0,0."""
        vert = self.x
        horiz = self.y
        rad_angle = atan2(vert, horiz)
        return rad_angle

    def ang_deg(self):
        """Returns angle in degrees, assume point is vector from 0,0."""
        rad_angle = self.ang_rad()
        deg = rad_angle * 180.0 / pi
        return deg

    def axrad(self):
        """Returns a tuple of (ax, rad) for plotting."""
        return (self.y, self.x)

    def radax(self):
        """Returns a tuple of (rad, ax) for checking if inside."""
        return (self.x, self.y)

    def rot_ccw_deg(self, ang):
        """Rotates the current vector by ccw degrees about 0,0. Returns self.

        Args:
            ang (float): angle or rotation in degrees, counter-clockwise (ccw)
        """
        ang = radians(ang)
        axial = self.y*cos(ang) - self.x*sin(ang)
        rad = self.y*sin(ang) + self.x*cos(ang)
        axial = round(axial, 5)
        rad = round(rad, 5)
        (self.x, self.y) = (rad, axial)
        return self

    def ang_bet_rad(self, other):
        """Returns angle between self and other vector in radians.

        Assumes vectors start at 0,0

        Args:
            other (Point): passed vector

        Returns:
            angle (float): radians between self vector and other vector
        """
        avect = Point(self.y*other.x - self.x*other.y, self.y*other.y + self.x*other.x)
        ang = avect.ang_rad()
        return ang

    def ang_bet_deg(self, other):
        """Returns angle between self and other vector in degrees.

        Assumes vectors start at 0,0

        Args:
            other (Point): passed vector

        Returns:
            angle (float): degrees between self vector and other vector
        """
        avect = Point(self.y*other.x - self.x*other.y, self.y*other.y + self.x*other.x)
        ang = avect.ang_deg()
        return ang

    def __str__(self):
        """Returns string listing object type, id number, and coordinates"""
        val = 'Point %s, (x, y)=(%.3f, %.3f)' % (self.get_name(), self.x, self.y)
        return val

    def set_esize(self, size):
        self.esize = size



class Line(base_classes.Idobj):
    """Stores a line from points start_pt to end_pt.

    Args:
        start_pt (Point): first point
        end_pt (Point): second point

    Attributes:
        points (list of Point) : list of the line's points [start_pt, end_pt]
        ediv (None or float): number of elements on the line
        midpt (Point): a mid-point between start_pt and end_pt
        nodes (list of Node): a list of meshed nodes on the line
        edge (bool): True if line is a non-shared edge in an area
        signlines (list): list of SignLine that use this Line
    """

    def __init__(self, start_pt, end_pt):
        self.points = [start_pt, end_pt]
        self.midpt = self.mid()
        self.ediv = None
        self.nodes = []
        self.edge = True
        self.signlines = []
        base_classes.Idobj.__init__(self)

    def reverse(self):
        """Reverses self."""
        self.points.reverse()

    @property
    def allpoints(self):
        """Returns all line defining points."""
        return self.points

    @property
    def areas(self):
        """Returns all areas that signlines use."""
        areas = set()
        for sline in self.signlines:
            areas.add(sline.parent)
        return list(areas)

    def __eq__(self, other):
        """Returns bool telling if self == other.

        Returns True if the lines are the same type, and contain the same
        list of points.
        """
        if isinstance(other, self.__class__):
            mypts = self.points
            otherpts = other.points
            same = True
            for (mypt, otherpt) in zip(mypts, otherpts):
                if mypt != otherpt:
                    same = False
                    break
            return same
        else:
            return False

    def __hash__(self):
        """Set the hash of the item to the id number.

        This allows one to make sets of these items.
        """
        return self.id

    def add_signline(self, signline):
        """Adds a signline to self.signlines."""
        if signline not in self.signlines:
            self.signlines.append(signline)

    def get_name(self):
        """Returns the line name."""
        return 'L'+str(self.id)

    def length(self):
        """Returns the length of the line."""
        return (self.pt(0)-self.pt(1)).length()

    def save_to_points(self):
        """This stores this line in the line's child points."""
        for point in self.allpoints:
            point.set_line(self)

    def set_ediv(self, ediv):
        """Sets the number of element divisions on the line when meshing.

        Args:
            ediv (int): number of required elements on this line
        """
        self.ediv = ediv

    def set_esize(self, esize):
        """Sets the size of mesh elements on the line when meshing.

        Args:
            esize (float): size of mesh elements on this line
        """
        for i in range(2):
            self.pt(i).set_esize(esize)

    def signed_copy(self, sign):
        """Returns a SignLine copy of this Line with the passed sign."""
        return SignLine(self, sign)

    def pt(self, ind):
        """Returns the start or end point of the line.

        Args:
            ind (int): index of the point we want, 0=start, 1=end

        Returns:
            Point: requested Point
        """
        return self.points[ind]

    def set_pt(self, ind, point):
        """Update the line's point at ind to the new point.

        Args:
            ind (int): index of the point we're updating, 0=start, 1=end
            point (Point): new point we are assigning to Line
        """
        # remove this line from the old point
        self.points[ind].lines.remove(self)
        # update the point
        self.points[ind] = point
        self.midpt = self.mid()
        self.save_to_points()
        other_point = self.pt(int(not ind))
        if point == other_point:
            # we are making a zero length line
            other_point.lines.remove(self)
            if self.id != -1:
                if len(self.signlines) > 0:
                    area = self.signlines[0].lineloop.parent
                    area.part.fea.lines.remove(self)
            for sline in self.signlines:
                sline.lineloop.remove(sline)

    def mid(self):
        """Calculates and returns a midpoint between start and end points."""
        point = (self.pt(0) + self.pt(1))*0.5
        return point

    def touches(self, other):
        """Checks if a line is connected to a passed line.

        Args:
            other (Line or Arc): passed other line

        Returns:
            True: beginning or end point connect to other line
            False: lines are not connected
        """
        [start_pt, end_pt] = [self.pt(0), self.pt(1)]
        if start_pt in other.points or end_pt in other.points:
            return True
        else:
            return False

    def offset(self, dist):
        """Returns a new line offset from the dirrent line by value dist.

        Right hand rule is followed assuming x is the origina line.
        Z points out of page.
        Y is the direction of line offset.

        Args:
            dist (float): distance to offset the new line by

        Returns:
            tmpline (Line): new Line which is offset from self by dist
        """
        lvect = self.get_perp_vec()
        lvect.make_unit()
        # scale it by the new amount
        lvect = lvect*dist
        # add the vector onto the defining points
        start_pt = self.pt(0) + lvect
        end_pt = self.pt(1) + lvect
        tmpline = Line(start_pt, end_pt)
        return tmpline

    def get_abc(self):
        """ Returns a list of abc terms for a line definition.

        ax + by + c = 0

        Returns:
            [a, b, c]: list of the above line terms
        """
        delta = self.pt(1)-self.pt(0)
        dx_ = delta.x
        dy_ = delta.y
        (a__, b__, c__) = (0.0, 0.0, 0.0)
        if dy_ == 0:
            # vertical radial line
            # y = 5 --> 0 = -y + 5 --> 0 = 0x -y + 5
            (b__, c__) = (-1.0, self.pt(0).y)
        elif dx_ == 0:
            # horizontal axial line
            # x = 5 --> 0 = -x + 5 --> 0 = -x +0y + 5
            (a__, c__) = (-1.0, self.pt(0).x)
        else:
            # y = ax + c --> 0 = ax - y + c
            slope = dy_/dx_
            offset = self.pt(0).y - slope*self.pt(0).x
            (a__, b__, c__) = (slope, -1.0, offset)
        return [a__, b__, c__]

    def get_perp_vec(self, point=None):
        """ Returns vector perpendicular to current line.

        Vector is created by rotating the current line ccw 90 degrees
        around the start point.

        Args:
            point (Point): defaults to None, unused for Line, used for Arc

        Returns:
            lvect (Point): vector that is perpendiculat to the current line
        """
        lvect = self.pt(1) - self.pt(0)
        lvect.rot_ccw_deg(90)
        return lvect

    def get_tan_vec(self, point):
        """ Returns vector tangent to the current line.

        Vector points from the passed point to a unit location further away.

        Args:
            point (Point): start point

        Returns:
            lvect (Point): vector that is tangent to the current line
        """
        end = self.pt(0)
        if point == end:
            end = self.pt(1)
        lvect = end - point
        lvect.make_unit()
        return lvect

    def arc_tang_intersection(self, point, mag):
        """Returns an intersection point on this line of an arc at center point

        Args:
            point (Point): arc center point
            mag (float): passed radius distance from the arc center to the line

        Returns:
            newpt (Point or None):
                Point: the intersection point
                None: returns None if no intersection point exists
        """
        vect = self.get_perp_vec()
        vect.make_unit()
        pt2 = point + vect*(-2*mag)

        tmpline = Line(point, pt2)
        newpt = tmpline.intersects(self)
        if type(newpt) == type(None):
            print('Intersection failed!')
        return newpt

    def coincident(self, point):
        """Checks to see if point is on the line.

        Args:
            point (Point): input point to check

        Returns:
            bool: True if point on line or False if not
        """
        # check if point is almost equal to end points
        dist = point - self.pt(0), point - self.pt(1)
        dist = [item.length() for item in dist]
        if min(dist) < ACC:
            return True
        # check if point is on the lines
        [a__, b__, c__] = self.get_abc()
        remainder = abs(a__*point.x + b__*point.y + c__)
        # print('  coincident: %s %s' % (self.get_name(), point))
        # print('  remainder: %.6f' % remainder)
        if remainder < ACC:
            # point is on the line equation, but is it between end pts
            xvals = [apoint.x for apoint in self.points]
            yvals = [apoint.y for apoint in self.points]
            xin = min(xvals)-ACC <= point.x <= max(xvals)+ACC
            yin = min(yvals)-ACC <= point.y <= max(yvals)+ACC
            if xin and yin:
                # we're on the line
                return True
            else:
                # we're off of the line
                # print('  outside of bounds!')
                return False
        else:
            # point is not on line equation
            return False

    def intersects(self, other):
        """Checks if other line intersects this line.

        Args:
            other (Line, Arc, Signline, SignArc): other line to check

        Returns:
            Point or None:
                Point: intersection point
                None: None if lines don't intersect
        """
        # other is cutline when checking
        if isinstance(other, Line) or isinstance(other, SignLine):
            terms_1 = self.get_abc()
            terms_2 = other.get_abc()
            p_1, p_2 = self.pt(0), self.pt(1)
            p_3, p_4 = other.pt(0), other.pt(1)
            dist = [p_1 - p_3, p_1 - p_4, p_2 - p_3, p_2 - p_4]
            vals = [p_1, p_1, p_2, p_2]
            dist = [item.length() for item in dist]
            # check each of self's point to see if it == other's points
            for (distance, val) in zip(dist, vals):
                if distance < ACC:
                    return Point(val.x, val.y)
            if terms_1[:2] == terms_2[:2]:
                # lines are parallel and will not have an intersection point
                #print('Line-Line intersection = None: same slope')
                return None
            denom = (p_1.x-p_2.x)*(p_3.y-p_4.y) - (p_1.y-p_2.y)*(p_3.x-p_4.x)
            if abs(denom) < ACC:
                # lines are nearly identical slope
                #print('Line-Line intersection = None because of denom')
                return None
            else:
                # intersection calculation
                a_array = array([[terms_1[0], terms_1[1]],
                                 [terms_2[0], terms_2[1]]])
                b_array = array([-terms_1[2], -terms_2[2]])
                result = linsolve(a_array, b_array)
                point = Point(result[0], result[1])
                # check against existing points on self
                dist = [point - p_1, point - p_2]
                vals = [p_1, p_2]
                dist = [item.length() for item in dist]
                for (distance, loc) in zip(dist, vals):
                    if distance < ACC:
                        point = Point(loc.x, loc.y)
                        break
                on_self = self.coincident(point)
                on_other = other.coincident(point)
                if on_self and on_other:
                    """
                    s_name = self.get_name()
                    o_name = other.get_name()
                    print('X Checking intersection of %s and %s' % (s_name, o_name))
                    print('  %s' % point)
                    print('  (on_self, on_other): (%s, %s)' % (on_self, on_other))
                    """
                    return point
                else:
                    """
                    print('Line-Line intersection = None point not on lines')
                    print(' %s' % point)
                    print(' on_self: %s' % on_self)
                    print(' on_other: %s' % on_other)
                    """
                    return None
        elif isinstance(other, Arc) or isinstance(other, SignArc):
            # arc line intersection
            return other.instersects(self)

    def __str__(self):
        """Returns string listing object type, name, and points"""
        start = self.pt(0)
        end = self.pt(1)
        val = 'Line %s, start: %s end: %s' % (self.get_name(), start, end)
        return val

class SignLine(Line, base_classes.Idobj):
    """Makes a signed line.

    Attributes:
        line (Line): parent line
        sign (int): 1 = positive, or -1 = negative
        lineloop (None or LineLoop): parent LineLoop, default is None
        faces (list): list of element faces
        n1 (list): list of first order nodes on line
        nodes (list): nodes on the line
    """
    def __init__(self, parent_line, sign):
        self.line = parent_line
        self.sign = sign
        self.lineloop = None
        self.faces = []
        self.n1 = []
        base_classes.Idobj.__init__(self)

    def pt(self, index):
        """Returns the point at the start or end of the line.

        Args:
            index (int): 0 = start point, 1 = end point
        """
        if self.sign == -1:
            return self.line.points[not index]
        else:
            return self.line.points[index]

    def set_pt(self, index, point):
        """Sets the start or end point of the line to a new point.

        Args:
            index (int): 0 for start point, 1 for end point
            point (Point): the new point we are updating to
        """
        self.line.set_pt(index, point)

    def set_ediv(self, ediv):
        """Applies the element divisions onto the parent line."""
        self.line.set_ediv(ediv)

    def set_esize(self, esize):
        """Applies the element size onto the parent line."""
        self.line.set_esize(esize)

    def set_lineloop(self, lineloop):
        """Sets the parent LineLoop"""
        # this is needed to cascade set ediv up to FEA model and down onto
        # other lines with this ediv
        self.lineloop = lineloop

    def get_name(self):
        """Returns the SignLine name."""
        if self.sign == -1:
            return '-L'+str(self.line.id)
        else:
            return 'L'+str(self.line.id)

    def label(self, ax):
        """Labels the line on the matplotlib ax axis.

        Args:
            ax (Matplotlib axis): Matplotlib axis
        """
        # find angle of perpendiclar line so we can set text alignment
        lvect = self.get_perp_vec()
        ang = lvect.ang_deg()
        horiz, vert = get_text_hv(ang)
        axial = self.midpt.y
        radial = self.midpt.x
        ax.text(axial, radial, self.get_name(), ha=horiz, va=vert)

    def plot(self, ax, label=True):
        """Draws and labels the line onto the passed Matplotlib axis.

        Args:
            ax (Matlotlib axis): Matplotlib axis
            label (bool): if True label the line
        """
        lax = [pt.y for pt in self.points]
        lrad = [pt.x for pt in self.points]
        # draw line
        ax.plot(lax, lrad)
        if label:
            self.label(ax)

    def signed_copy(self, sign):
        """Returns a SignLine copy of this line."""
        realsign = self.sign*sign
        return SignLine(self.line, realsign)

    @property
    def nodes(self):
        """Gets the line nodes."""
        return self.line.nodes

    @nodes.setter
    def nodes(self, nlist):
        """Sets the line nodes."""
        self.line.nodes = nlist

    @property
    def midpt(self):
        """Returns the mid point."""
        return self.line.midpt

    @property
    def edge(self):
        """Returns a bool saying if this is an edge line."""
        return self.line.edge

    @edge.setter
    def edge(self, isedge):
        """Sets the line nodes."""
        self.line.edge = isedge

    @property
    def points(self):
        """Returns a list of the line's points [pstart, pend]"""
        pstart = self.pt(0)
        pend = self.pt(1)
        return [pstart, pend]

    def reverse(self):
        """Reverses self.line"""
        self.line.reverse()


class Arc(base_classes.Idobj):
    """Makes an arc from points start_pt to end_pt about center actr.

    Arcs should be 90 degrees maximum.

    Arguments:
        start_pt (Point): first point
        end_pt (Point): second point
        actr (Point): arc center

    Attributes:
        points (list of Point) : list of the line's points [start_pt, end_pt]
        allpoints (list of Point): [start_pt, end_pt, actr]
        actr (Point): arc center
        radius (float): radius of the arc
        concavity (str):

            - 'concave': concave arc
            - 'convex': convex arc

        midpt (Point): a mid-point between p1 and p2
        ediv (None or int): number of elements on arc
        nodes (list of Node): a list of meshed nodes on the line
        edge (bool): True if arc is a non-shared edge in an area
        signlines (list): list of SignArc that use this Arc
    """

    def __init__(self, start_pt, end_pt, actr):
        self.points = [start_pt, end_pt]
        self.actr = actr
        self.actr.arc_center = True
        self.radius = (start_pt - actr).length()
        self.concavity = self.get_concavity()
        self.midpt = self.mid()
        self.ediv = None
        self.nodes = []
        self.edge = True
        self.signlines = []
        base_classes.Idobj.__init__(self)

    def reverse(self):
        """Reverses self."""
        self.points.reverse()
        self.concavity = self.get_concavity()

    @property
    def allpoints(self):
        """Returns all points including arc center"""
        return self.points + [self.actr]

    @property
    def areas(self):
        """Returns all areas that signlines use."""
        areas = set()
        for sline in self.signlines:
            areas.add(sline.parent)
        return list(areas)

    def __eq__(self, other):
        """Returns bool telling if self == other

        Returns True if the lines are the same type, and contain the same
        list of points.
        """
        if isinstance(other, self.__class__):
            # check to see if the end points are the same
            mypts = self.points
            otherpts = other.points
            same = True
            for (mypt, otherpt) in zip(mypts, otherpts):
                if mypt != otherpt:
                    same = False
                    break
            # flip to false if arc centers are different
            myactr = self.actr
            otheractr = other.actr
            if myactr != otheractr:
                same = False
            return same
        else:
            return False

    def __hash__(self):
        """Set the hash of the item to the id number.

        This allows one to make sets of these items.
        """
        return self.id

    def add_signline(self, signline):
        """Adds a signline to self.signlines."""
        if signline not in self.signlines:
            self.signlines.append(signline)

    def get_name(self):
        """Returns the arc name."""
        return 'L'+str(self.id)

    def length(self):
        """Returns the length of the arc."""
        a = self.pt(0)-self.actr
        b = self.pt(1)-self.actr
        ang = abs(a.ang_bet_rad(b))
        res = self.radius*ang
        return res

    def save_to_points(self):
        """This stores this line in the line's child points."""
        for point in self.allpoints:
            point.set_line(self)

    def set_ediv(self, ediv):
        """Sets the number of element divisions on the arc when meshing.

        Args:
            ediv (int): number of required elements on this arc
        """
        self.ediv = ediv

    def set_esize(self, esize):
        """Sets the size of mesh elements on the arc when meshing.

        Args:
            esize (float): size of mesh elements on this arc
        """

        for i in range(2):
            self.pt(i).set_esize(esize)

    def signed_copy(self, sign):
        """Returns a SignArc instance of this Arc with the passed sign."""
        return SignArc(self, sign)

    def pt(self, ind):
        """Returns the start or end point of the arc.

        Args:
            ind (int): index of the point we want, 0=start, 1=end

        Returns:
            Point: requested Point
        """
        return self.points[ind]

    def set_pt(self, ind, point):
        """Update the arc's point at ind to the new point.

        Args:
            ind (int): index of the point we're updating, 0=start, 1=end
            point (Point): new point we are assigning to Line
        """
        # remove this line from the old point
        self.points[ind].lines.remove(self)
        # set the new point
        self.points[ind] = point
        self.midpt = self.mid()
        self.save_to_points()
        other_point = self.pt(int(not ind))
        if point == other_point:
            other_point.lines.remove(self)
            for sline in self.signlines:
                sline.lineloop.remove(sline)
            if self.id != -1:
                if len(self.signlines) > 0:
                    area = self.signlines[0].lineloop.parent
                    area.part.fea.lines.remove(self)

    def mid(self):
        """Caclulates and returns the midpoint of the arc."""
        pt = self.get_pt_at(0.5)
        return pt

    def touches(self, other):
        """Checks if this arc is connected to a passed line.

        Args:
            other (Line or Arc): passed other line

        Returns:
            True: beginning or end point connect to other line
            False: lines are not connected
        """
        [p1, p2] = [self.pt(0), self.pt(1)]
        if p1 in other.points or p2 in other.points:
            return True
        else:
            return False

    @staticmethod
    def sgn(val):
        """Returns sign of the passed value.

        Helper function for self.interects code.

        Args:
            val (float): passed value

        Returns:
            1 or -1: sign of the passed value, assumes sign of zero = 1
        """
        if val < 0:
            return -1
        else:
            return 1

    def get_ang(self):
        """ Returns angle from beginning to end of arc from arc ctr in degrees

        Notes:
            Answer is between 180 and -180
            Positive is concave
            Negative is convex
        """
        a = self.pt(0)-self.actr
        b = self.pt(1)-self.actr
        ang = a.ang_bet_deg(b)
        return ang

    def get_ang_rad(self):
        """ Returns angle from beginning to end of arc from arc ctr in rad

        Notes:
            Answer is between pi and -pi
            Positive is concave
            Negative is convex
        """
        a = self.pt(0)-self.actr
        b = self.pt(1)-self.actr
        ang = a.ang_bet_rad(b)
        return ang

    def get_concavity(self, clockwise=True):
        """Returns concavity of this arc, 'concave' or 'convex'."""
        ang = self.get_ang()
        res = ''
        if clockwise:
            res = 'concave'
            if ang < 0:
                res = 'convex'
        else:
            res = 'convex'
            if ang < 0:
                res = 'concave'
        return res

    def get_pt_at(self, nondim):
        """Returns point on arc between start and end point at nondim dist.

        Args:
            nondim (float): distance between start and end point, 0=start, 1=end

        Returns:
            res (Point): requested point on arc
        """
        a = self.pt(0)-self.actr

        # neg is convex, pos is concave
        ang = self.get_ang()

        ang_delta = nondim*ang
        a.rot_ccw_deg(ang_delta)

        res = self.actr + a
        return res

    def get_perp_vec(self, pt):
        """ Returns vector perpendicular to current arc.

        Right hand rule is used:
            X+ is the direction of the arc from start to end.
            Z+ is out of the page
            Y+ is the direction of the perpendicular vector

        Args:
            pt (Point): point on arc where we want the perpendicular vector

        Returns:
            resv (Point): perpendicular vector
        """
        resv = None
        if self.concavity == 'convex':
            resv = pt - self.actr
        elif self.concavity == 'concave':
            resv = self.actr - pt
        return resv

    def get_tan_vec(self, point):
        """ Returns vector tangent to the current arc.

        Vector points from the passed point to a unit location further away.

        Args:
            point (Point): start point

        Returns:
            resv (Point): vector that is tangent to the current line
        """
        if self.concavity == 'concave':
            resv = self.actr - point
        elif self.concavity == 'convex':
            resv = point - self.actr
        if point == self.pt(0):
            resv.rot_ccw_deg(-90)
        else:
            resv.rot_ccw_deg(90)
        resv.make_unit()
        return resv

    def get_verts_codes(self, plot=True):
        """Returns a list of [verts, codes] vertices for plotting."""
        # http://www.spaceroots.org/documents/ellipse/node19.html
        # http://www.tinaja.com/glib/bezarc1.pdf
        # http://stackoverflow.com/questions/734076/geometrical-arc-to-bezier-curve

        # find the angle of rotation to the middle of the arc
        ang_offset = self.get_pt_at(0.5) - self.actr
        ang_offset = ang_offset.ang_deg()

        # store the half angle as theta
        theta = radians(abs(self.get_ang()))*0.5

        ax1 = (4 - cos(theta))/3
        rad1 = ((1 - cos(theta))*(cos(theta) - 3))/(3*sin(theta))
        pt1 = Point(rad1, ax1)*self.radius
        pt2 = Point(-rad1, ax1)*self.radius

        # fix the angles
        pt1.rot_ccw_deg(ang_offset)
        pt2.rot_ccw_deg(ang_offset)

        # fix the location
        pt1 += self.actr
        pt2 += self.actr

        # need to change the order if the arc is concave or convex
        verts = [pt1, pt2, self.pt(1)]
        if self.concavity == 'convex':
            verts = [pt2, pt1, self.pt(1)]

        if plot:
            verts = [point.axrad() for point in verts]
        else:
            verts = [point.radax() for point in verts]
        codes = [Path.CURVE4, Path.CURVE4, Path.CURVE4]
        return [verts, codes]

    def coincident(self, pt):
        """Checks to see if pt is on the arc.

        Args:
            pt (Point): input point to check

        Returns:
            bool: True if pt on arc or False if not
        """
        # (x - xc)^2 + (y - yc)^2 - r^2 = 0
        dist = pt - self.pt(0), pt - self.pt(1)
        dist = [item.length() for item in dist]
        if min(dist) < ACC:
            # True if point almost equal to end points
            return True
        remainder = abs((pt.x - self.actr.x)**2 + (pt.y - self.actr.y)**2
                        - self.radius**2)
        if remainder < ACC:
            # point is on the circle, but is it in the arc?
            # check if the angle traversed to the point is within the ang
            # traversed by the arc
            a1 = self.pt(0)-self.actr
            a2 = pt-self.actr
            ang_pt = a1.ang_bet_deg(a2)

            # this is the angle traversed byt the arc
            ang = self.get_ang()
            minval = min(0, ang)
            maxval = max(0, ang)
            if minval <= ang_pt <= maxval:
                # point is on our arc
                return True
            else:
                # point on circle but outside our arc
                return False
        else:
            # point not on arc
            return False

    def intersects(self, other):
        """Checks if other line intersects this arc.

        Args:
            other (Line or Arc): other line to check

        Returns:
            Point or None:
                Point: intersection point
                None: None if lines don't intersect
        """
        if isinstance(other, Line) or isinstance(other, SignLine):
            #print('|C Checking arc-line intersection')
            #print(' %s' % self)
            # arc-line intersection
            # reset the math to be about the circle centroid
            # http://mathworld.wolfram.com/Circle-LineIntersection.html
            pt1 = other.pt(0) - self.actr
            pt2 = other.pt(1) - self.actr
            # for us x and y are reversed because of y axial
            (x1, y1, x2, y2) = (pt1.y, pt1.x, pt2.y, pt2.x)
            dx = x2 - x1
            dy = y2 - y1
            rad = self.radius
            dr = (dx**2 + dy**2)**(0.5)
            D = x1*y2 - x2*y1
            dist = [self.pt(0) - other.pt(0), self.pt(0) - other.pt(1),
                    self.pt(1) - other.pt(0), self.pt(1) - other.pt(1)]
            vals = [self.pt(0), self.pt(0), self.pt(1), self.pt(1)]
            dist = [item.length() for item in dist]
            for (distance, val) in zip(dist, vals):
                if distance < ACC:
                    return Point(val.x, val.y)
            discr = 1.0*(rad**2)*(dr**2) - D**2
            if discr < 0:
                # no intersection
                #print('none returned because of discriminant')
                return None
            else:
                # tangent or intersection
                rsq = rad**2
                drsq = dr**2
                x1 = (D*dy + self.sgn(dy)*dx*(rsq*drsq-D**2)**(0.5))/drsq
                y1 = (-D*dx + abs(dy)*(rsq*(dr**2)-D**2)**(0.5))/drsq
                x2 = (D*dy - self.sgn(dy)*dx*(rsq*drsq-D**2)**(0.5))/drsq
                y2 = (-D*dx - abs(dy)*(rsq*drsq-D**2)**(0.5))/drsq

                # convert result points back to global coordinates
                pt1 = Point(y1, x1) + self.actr
                pt2 = Point(y2, x2) + self.actr
                res = []
                # check that the resultant points are on line and arcs
                if self.coincident(pt1) and other.coincident(pt1):
                    for apoint in self.points:
                        dist = pt1 - apoint
                        dist = dist.length()
                        if dist < ACC:
                            pt1 = Point(apoint.x, apoint.y)
                            break
                    res.append(pt1)
                if pt1 != pt2 and self.coincident(pt2) and other.coincident(pt2):
                    for apoint in self.points:
                        dist = pt2 - apoint
                        dist = dist.length()
                        if dist < ACC:
                            pt2 = Point(apoint.x, apoint.y)
                            break
                    res.append(pt2)
                if len(res) == 1:
                    return res[0]
                elif len(res) > 1:
                    return res
                elif len(res) == 0:
                    #print('none returned because no point from math')
                    return None
        elif isinstance(other, Arc):
            # arc-arc intersection
            # I have not yet written the code here
            return None

    def __str__(self):
        """Returns string listing object type, name, and points"""
        name = self.get_name()
        start, end = self.pt(0), self.pt(1)
        actr = self.actr
        val = 'Arc %s, start: %s end: %s center: %s' % (name, start, end, actr)
        return val

class SignArc(Arc, base_classes.Idobj):
    """Makes a signed arc.

    Attributes:
        line (Arc): parent arc
        sign (int): 1 = positive, or -1 = negative
        lineloop (None or LineLoop): parent LineLoop, default is None
        nodes (list): nodes on the arc
        faces (list): list of element faces
        n1 (list): list of first order nodes on line
    """

    def __init__(self, parent_line, sign):
        self.line = parent_line
        self.sign = sign
        self.lineloop = None
        self.faces = []
        self.n1 = []
        base_classes.Idobj.__init__(self)

    def pt(self, ind):
        """Returns the point at the start or end of the line."""
        if self.sign == -1:
            return self.line.points[not ind]
        else:
            return self.line.points[ind]

    def set_pt(self, ind, point):
        """Sets the passed point to at the index location."""
        self.line.set_pt(ind, point)

    def set_ediv(self, ediv):
        """Apply the element divisions onto the parent line."""
        self.line.set_ediv(ediv)

    def set_esize(self, esize):
        """Applies the element size onto the parent line."""
        self.line.set_esize(esize)

    def set_lineloop(self, lineloop):
        """Sets the parent LineLoop"""
        # this is needed to cascade set ediv up to FEA model and down onto
        # other lines with this ediv
        self.lineloop = lineloop

    def get_name(self):
        """Returns the line name."""
        if self.sign == -1:
            return '-L'+str(self.line.id)
        else:
            return 'L'+str(self.line.id)

    def label(self, ax):
        """Labels the arc on the matplotlib ax axis.

        Args:
            ax (Matplotlib axis): Matplotlib axis
        """
        lvect = self.get_perp_vec(self.midpt)
        ang = lvect.ang_deg()
        horiz, vert = get_text_hv(ang)
        (axial, radial) = (self.midpt.y, self.midpt.x)
        ax.text(axial, radial, self.get_name(), ha=horiz, va=vert)

    def plot(self, ax, label=True):
        """Draws and labels the arc onto the passed Matplotlib axis.

        Args:
            ax: Matplotlib axis
            label (bool): if True, label the arc
        """
        ctr = self.actr
        vect1 = None
        sign = 1
        if self.concavity == 'concave':
            vect1 = self.pt(0)-ctr
        else:
            sign = -1
            vect1 = self.pt(1)-ctr
        rad = self.radius
        ang1 = (vect1.ang_deg())
        ang2 = ang1 + sign*self.get_ang()

        # matplotlib assumes ccw arc drawing, calculix assumes cw drawing
        a = AArc(xy=[ctr.y, ctr.x], width=2*rad, height=2*rad, angle=0,
                 theta1=ang1, theta2=ang2)
        ax.add_artist(a)
        if label:
            self.label(ax)

    def signed_copy(self, sign):
        """Returns a SignArc copy of this arc."""
        realsign = self.sign*sign
        return SignArc(self.line, realsign)

    @property
    def nodes(self):
        """Gets the line nodes."""
        return self.line.nodes

    @nodes.setter
    def nodes(self, nlist):
        """Sets the line nodes."""
        self.line.nodes = nlist

    @property
    def actr(self):
        """Return the arc center."""
        return self.line.actr

    @property
    def radius(self):
        """Return the arc radius."""
        return self.line.radius

    @property
    def midpt(self):
        """Return the mid point."""
        return self.line.midpt

    @property
    def edge(self):
        """Returns a bool saying if this is an edge line."""
        return self.line.edge

    @edge.setter
    def edge(self, isedge):
        """Sets the line nodes."""
        self.line.edge = isedge

    @property
    def points(self):
        """Return a list of the arc's points [pstart, pend]"""
        pstart = self.pt(0)
        pend = self.pt(1)
        return [pstart, pend]

    @property
    def concavity(self):
        """Return the arc's concavity 'concave' or 'convex'"""
        if self.sign == 1:
            return self.line.concavity
        else:
            if self.line.concavity == 'concave':
                return 'convex'
            else:
                return 'concave'

    def reverse(self):
        """Reverses self.line"""
        self.line.reverse()


class LineLoop(list):
    """Makes a LineLoop, which stores multiple Line/Arc or SignLine/SignArc.

    Args:
        items (list): Line or Arc or SignLine or SignArc
        hole_bool (bool): True if hole, False otherwise
        parent (Area or None): parent Area, defaults to None

    Attributes:
        closed (bool): True if closed
        area (float): the loop's area, can be pos or neg. Pos = cw, neg = ccw
        hole (bool): True if loop is a hole
        ccw (bool): True if loop is counter-clockwise
        center (Point): a point at the loop centroid
        parent (Area or None): the loop's parent Area
    """

    def __init__(self, items=[], hole_bool=False, parent=None):
        super().__init__(items)
        self.id = -1
        for item in self:
            if isinstance(item, SignLine) or isinstance(item, SignArc):
                item.lineloop = self
                if item not in item.line.signlines:
                    item.line.add_signline(item)
        self.hole = hole_bool
        self.parent = parent

    def set_id(self, id):
        """Sets the id number"""
        self.id = id

    def __hash__(self):
        """Sets the hash of the item to the id number.

        This allows one to make sets of these items.
        """
        return self.id

    @property
    def closed(self):
        """Returns True if Closed, False if not."""
        if len(self) < 2:
            return False
        else:
            for ind, curr_line in enumerate(self):
                lower_ind = ind-1
                prev_line = self[lower_ind]
                if curr_line.pt(0) != prev_line.pt(1):
                    return False
            return True

    @property
    def center(self):
        """Returns None or a Point at the line loop's center."""
        if self.closed == False:
            return None
        else:
            cx_sum = 0
            cy_sum = 0
            a_sum = 0
            arcs = []
            for item in self:
                if isinstance(item, Line) or isinstance(item, SignLine):
                    start, end = item.pt(0), item.pt(1)
                    det_val = det([[start.x, end.x], [start.y, end.y]])
                    cx_sum += (start.x + end.x)*det_val
                    cy_sum += (start.y + end.y)*det_val
                    a_sum += det_val
                elif isinstance(item, Arc) or isinstance(item, SignArc):
                    # concave: draw to actr, adder is neg
                    # convex:  draw to actr, adder is pos
                    tmp_list = [[item.pt(0), item.actr],
                                [item.actr, item.pt(1)]]
                    for [start, end] in tmp_list:
                        det_val = det([[start.x, end.x], [start.y, end.y]])
                        cx_sum += (start.x + end.x)*det_val
                        cy_sum += (start.y + end.y)*det_val
                        a_sum += det_val
                    arcs.append(item)
            area_val = a_sum*0.5
            val_list = []
            if area_val != 0:
                cx_val = (1/(6*area_val))*cx_sum
                cy_val = (1/(6*area_val))*cy_sum
                val_list.append([area_val, cx_val, cy_val])
            for arc in arcs:
                ang = arc.get_ang_rad()
                area_val = -0.5*ang*(arc.radius**2)
                half_ang = abs(ang)*0.5
                vect = arc.midpt - arc.actr
                vect.make_unit()
                mag = 2*arc.radius*sin(half_ang)/(3*half_ang)
                vect = vect*mag
                center = arc.actr + vect
                val_list.append([area_val, center.x, center.y])
            a_sum = sum([aval for [aval, cx, cy] in val_list])
            cxa_sum = sum([cx*aval for [aval, cx, cy] in val_list])
            cya_sum = sum([cy*aval for [aval, cx, cy] in val_list])
            cx_val = cxa_sum/a_sum
            cy_val = cya_sum/a_sum
            return Point(cx_val, cy_val)

    @property
    def ccw(self):
        """Returns bool telling if the area is closed and clockwise."""
        if self.closed == False:
            return False
        else:
            area_val = self.area
            if area_val < 0:
                return True
            else:
                return False

    @property
    def area(self):
        """Returns the loop's signed area.

        Returns 0 if not closed.
        Pos = cw, neg = ccw

        Formula from: http://mathworld.wolfram.com/PolygonArea.html
        """
        res = 0
        adder = 0
        arcs = []
        if self.closed == False:
            return res
        else:
            for item in self:
                if isinstance(item, Line) or isinstance(item, SignLine):
                    start, end = item.pt(0), item.pt(1)
                    det_matrix = [[start.x, end.x], [start.y, end.y]]
                    res += det(det_matrix)
                elif isinstance(item, Arc) or isinstance(item, SignArc):
                    # concave: draw to actr, adder is neg
                    # convex:  draw to actr, adder is pos
                    start, end = item.pt(0), item.actr
                    det_matrix = [[start.x, end.x], [start.y, end.y]]
                    res += det(det_matrix)
                    start, end = item.actr, item.pt(1)
                    det_matrix = [[start.x, end.x], [start.y, end.y]]
                    res += det(det_matrix)
                    arcs.append(item)
            res = 0.5*res
            # need to loop through arcs here, now that we know if cw or ccw
            for arc in arcs:
                ang = arc.get_ang_rad()
                tmp_adder = -0.5*ang*(arc.radius**2)
                adder += tmp_adder
            res = res + adder
            return res

    def reverse(self):
        """Reverse the loop direction cw<-->ccw

        This modifies the order of lines, and flips each line.
        """
        for line in self:
            line.reverse()
        super().reverse()

    def set_parent(self, parent):
        """Sets the line loop's parent part."""
        self.parent = parent

    def get_patch(self):
        """Returns a patch of the LineLoop, or None of not closed."""
        if self.closed:
            codes, verts = [], []
            # start poly
            codes += [Path.MOVETO]
            verts += [self[0].pt(0).radax()]
            for line in self:
                if isinstance(line, SignLine) or isinstance(line, Line):
                    verts.append(line.pt(1).radax())
                    codes.append(Path.LINETO)
                elif isinstance(line, SignArc) or isinstance(line, Arc):
                    lverts, lcodes = line.get_verts_codes(plot=False)
                    verts += lverts
                    codes += lcodes
                if line == self[-1]:
                    # close poly if at last line
                    verts.append((0, 0))
                    codes.append(Path.CLOSEPOLY)
            path = Path(verts, codes)
            patch = PathPatch(path, facecolor='yellow', linewidth=0.0)
            return patch
        return None

    def contains_point(self, point):
        """Returns bool telling if pt is inside the LineLoop.

        Args:
            pt (Point): point we are checking

        Returns:
            contains (bool): True if in this area False if not
        """
        contains = False
        if self.closed:
            patch = self.get_patch()
            contains = patch.contains_point(point.radax())
        return contains

    def inside(self, other):
        """Checks to see if self is inside other.

        Only checks self's points.
        """
        if self.closed == False:
            return False
        if isinstance(other, LineLoop):
            points = set()
            for sline in self:
                points.update(sline.points)
            patch = other.get_patch()
            for point in points:
                contains = patch.contains_point(point.radax())
                if contains == False:
                    return False
            return True
        else:
            return False

    def append(self, item):
        """Adds an item to this LineLoop

        Args:
          item (SingArc or SignLine): item to add to the list
        """
        if isinstance(item, SignLine) or isinstance(item, SignArc):
            item.set_lineloop(self)
        super().append(item)
        return item

    def remove(self, item):
        """Removes an SignLine or SignArc item from this LineLoop

        Args:
          item (SignArc or SignLine): item to remove
        """
        if isinstance(item, SignLine) or isinstance(item, SignArc):
            item.set_lineloop(None)
            if item not in item.line.signlines:
                print('''===THIS SLINE NOT IN SLINE.LINE.SIGNLINES!===''')
                print('Removing sline %s from loop in area %s'
                      % (item.get_name(), self.parent.get_name()))
                print('SLINE.LINE %s' % item.line)
                print('SLINE.LINE.SIGNLINES %r' % item.line.signlines)
        super().remove(item)

    def insert(self, index, item):
        """Inserts a SignLine or SignArc item into this LineLoop

        Args:
          item (SingArc or SignLine): item to insert
        """
        if isinstance(item, SignLine) or isinstance(item, SignArc):
            item.set_lineloop(self)
        super().insert(index, item)

    def __str__(self):
        """Prints the LineLoop definition."""
        res = 'Number of items %i\n' % len(self)
        for line in self:
            res += '%s\n' % line
        return res

class Area(base_classes.Idobj):
    """ Makes an area.

    Area is closed in a clockwise direction.

    Args:
        part (Part): parent Part
        line_list (list): list of Lines and Arcs that close the area

    Attributes:
        part (Part): parent part
        closed (boolean): True if closed, False if not
        exlines (LineLoop): LineLoop of signlines that define the area exterior
        signlines (list): list of signed lines or arcs that define the area
        lines (list): a list of all the lines that make the area, includes hole
            lines
        points (list): a list of all points making the area, excludes arc centers
        allpoints (list): a list of all points, includes arc centers
        holepoints (list): a list of hole points, excludes arc centers
        matl (Matl): the material fo the area
        etype (str): element type of area. Options are:
            'plstress': plane stress
            'plstrain': plane strain
            'axisym': axisymmetric
        area (double): in-plane part area (ballpark number, arcs not calc right)
        holes (list): list of LineLoop of signlines
        center (Point or None): the area centroid point. None before closed.
        nodes (list): child mesh nodes that are in the area
        elements (list): child elements that are in the area
    """

    def __init__(self, part, line_list=[]):
        base_classes.Idobj.__init__(self)
        self.part = part # parent part reference
        self.closed = False
        self.matl = None
        self.etype = None
        self.area = 0.0
        self.holes = []
        self.center = None
        self.nodes = []
        self.elements = []
        exloop = LineLoop(line_list, False, self)
        self.part.fea.register(exloop)
        self.exlines = exloop
        if self.exlines.closed == True:
            self.close()

    def __hash__(self):
        """Returns the item's id as its hash."""
        return self.id

    @property
    def lines(self):
        """Returns a list of all Line and Arc under this area."""
        mylist = []
        loops = [self.exlines] + self.holes
        for loop in loops:
            line_list = [line.line for line in loop]
            mylist += line_list
        return mylist

    @property
    def signlines(self):
        """Returns a list of all SignLine and SignArc under this area."""
        mylist = []
        loops = [self.exlines] + self.holes
        for loop in loops:
            mylist += loop
        return mylist

    @property
    def points(self):
        """Returns a list of all points under this area; excludes arc centers."""
        pts = set()
        for line in self.lines:
            pts.update(line.points)
        return list(pts)

    @property
    def allpoints(self):
        """Returns a list of all points under this area; includes arc centers."""
        pts = set()
        for line in self.lines:
            pts.update(line.allpoints)
        return list(pts)

    @property
    def holepoints(self):
        """Returns a list of all points in this area's holes."""
        pts = set()
        for hole in self.holes:
            for sline in hole:
                pts.update(sline.points)
        return list(pts)

    def __str__(self):
        """Returns a string listing the area's name."""
        return self.get_name()

    def get_name(self):
        """Returns the area name."""
        return 'A'+str(self.id)

    def close(self):
        """Closes the area."""
        # sets closed=True
        # calculates the area center
        self.closed = True
        if self.exlines.ccw == True:
            self.exlines.reverse()
        self.area, self.center = self.calc_area_center()

    def line_from_startpt(self, pt):
        """Returns the signline in the area that start at the passed pt.

        Args:
            pt (Point): the start point of the line one wants

        Returns:
            match (None or SignLine or SignArc):
            |  Returns None if no line is found, otherwise the correct line or
            |  arc which starts with pt will be
            |  returned.
        """
        # returns the line that starts on the given point
        matches = []
        area_slines = self.signlines
        for sline in area_slines:
            if sline.pt(0) == pt:
                matches.append(sline)
        if len(matches) == 1:
            return matches[0]
        elif len(matches) > 1:
            # we have a saw cut in the area and we're cutting on one of its pts
            print('>1 slines starting with %s' % pt.get_name())
            for match in matches:
                slines = set(match.line.signlines)
                if len(slines) == 2 and slines.issubset(area_slines):
                    # this is the saw cut one
                    pass
                else:
                    # we want the non saw cut one
                    return match
        return None

    def calc_area_center(self):
        """Calculates and returns the area and centroid Point.

        Returns:
            list: [area, Point]
        """
        loops = [self.exlines] + self.holes
        val_list = []
        for loop in loops:
            if loop.closed == True:
                aval, cval = loop.area, loop.center
                val_list.append([aval, cval])
        a_sum = sum([aval[0] for aval in val_list])
        cxa_sum = sum([center.x*aval for [aval, center] in val_list])
        cya_sum = sum([center.y*aval for [aval, center] in val_list])
        cx_val = cxa_sum/a_sum
        cy_val = cya_sum/a_sum
        center = Point(cx_val, cy_val)
        return [a_sum, center]

    def get_maxlength(self):
        """Returns max distance between points in an area."""
        pts = self.points
        maxlen = 0.0
        # loop through points checking dist to next point
        for ind, p1 in enumerate(pts[:-1]):
            for p2 in pts[ind:]:
                vect = p1 - p2
                dist = vect.length()
                if dist > maxlen:
                    maxlen = dist
        return maxlen

    def rem_sline(self, sline):
        """Removes a sline from this area."""
        loops = [self.exlines] + self.holes
        for loop in loops:
            if sline in loop:
                print('Line %s removed from %s'
                      % (sline.get_name(), self.get_name()))
                loop.remove(sline)
        if sline.id != -1:
            line = sline.line
            self.part.fea.signlines.remove(sline)
            if len(line.signlines) == 0:
                self.part.fea.lines.remove(line)

    def add_sline(self, signline):
        """Adds signline to the area definition.

        SignLine is added to the end of the lines list.
        Area is closed if needed.
        """
        # this adds a line to the area
        self.exlines.append(signline)
        if self.exlines.closed:
            self.close()

    def add_hole_sline(self, signline):
        """Adds signline to the area hole definition.

        Line is added to the end of the lines list.
        Area is closed if needed.

        Returns:
            closed (bool): boolean telling if the hole was closed
        """
        make_new = (len(self.holes) == 0)
        if make_new == False:
            if self.holes[-1].closed == True:
                make_new = True
        if make_new:
            hole = LineLoop([], hole_bool=True, parent=self)
            hole = self.part.fea.lineloops.append(hole)
            self.holes.append(hole)
        self.holes[-1].append(signline)
        # reverse the hole if it's closed in the cw direction
        if self.holes[-1].closed == True:
            if self.holes[-1].ccw == False:
                self.holes[-1].reverse()
            self.area, self.center = self.calc_area_center()
        return self.holes[-1].closed

    def get_patch(self):
        """Returns a patch of the area."""
        if self.closed:
            loops = [self.exlines] + self.holes
            codes, verts = [], []
            for loop in loops:
                # start poly
                codes += [Path.MOVETO]
                verts += [loop[0].pt(0).axrad()]
                for sline in loop:
                    if isinstance(sline, SignLine):
                        verts.append(sline.pt(1).axrad())
                        codes.append(Path.LINETO)
                    elif isinstance(sline, SignArc):
                        lverts, lcodes = sline.get_verts_codes()
                        verts += lverts
                        codes += lcodes
                    if sline == loop[-1]:
                        # close poly if at last line
                        verts.append((0, 0))
                        codes.append(Path.CLOSEPOLY)
            path = Path(verts, codes, _interpolation_steps=2)
            patch = PathPatch(path, linewidth=0.0)
            return patch

    def label(self, axis):
        """Labels the area on a Matplotlib axis

        Args:
            axis (Matplotlib Axis): Matplotlib Axis
        """
        axis.text(self.center.y, self.center.x, self.get_name(),
                  ha='center', va='center')

    def plot(self, axis, label=True, color='yellow'):
        """Plots the area on a Matplotlib axis.

        Args:
            axis (Matplotlib axis): Matplotlib axis
            label (bool): if True, label area
        """
        # http://stackoverflow.com/questions/8919719/how-to-plot-a-complex-polygon
        if self.closed:
            # plot area patch
            patch = self.get_patch()
            patch.set_color(color)
            axis.add_patch(patch)
            # label ares
            if label:
                self.label(axis)

    def contains_point(self, point):
        """Returns bool telling if pt is inside the area.

        Args:
            pt (Point): point we are checking

        Returns:
            contains (bool): True if in this area False if not
        """
        patch = self.get_patch()
        contains = patch.contains_point(point.axrad())
        return contains

    def line_insert(self, lgiven, lnew, after=True):
        """Inserts line lnew before or after the given line.

        Args:
            lgiven (SignLine or SignArc): the line we are inserting relative to
            lnew (SignLine or SignArc): the new line we will insert
            after (bool): if True insert lnew after lgiven, false insert before

        Returns:
            bool: True if succeeded, False if failed
        """
        if lgiven in self.signlines:
            print('Inserting line %s into area %s next to %s' %
                  (lnew.get_name(), self.get_name(), lgiven.get_name()))
            lineloop = lgiven.lineloop
            type = 'exterior lines'
            if lineloop.hole:
                type = 'hole lines'
            ind = lineloop.index(lgiven)
            if after:
                ind += 1
            print(' Inserting line into %s outer lines' % self.get_name())
            lineloop.insert(ind, lnew)
            return True
        else:
            print('ERROR: Passed line was not in %s!' % self.get_name())
            print('You must pass a line in this area!')
            return False

    def set_child_ccxtypes(self):
        """Assigns the child element ccx types."""
        if self.etype != None and self.elements != []:
            # if etype is set and elements have been assigned
            for element in self.elements:
                element.set_ccxtype(self.etype)

    def set_etype(self, etype):
        """Sets the area etype.

        Sets the ccxtype on elements if they have been assigned to this area.

        Args:
            etype (str): element type
                'plstress' plane stress
                'plstrain': plane strain
                'axisym': axisymmetric
        """
        self.etype = etype
        self.set_child_ccxtypes()

    def set_esize(self, esize):
        """Sets the size of mesh elements on the area when meshing.

        Args:
            esize (float): size of mesh elements on this area
        """
        #size if set if no mesh size on points is peresent
        for p in self.points:
            p.set_esize(esize)

    def update(self, line_list):
        """Updates the area's external lines to the passed list.

        If the area is cosed, self.close() will be called.

        Args:
            line_list (list): list of SignLine and SignArc items
        """
        to_rem = list(self.exlines)
        for sline in to_rem:
            self.exlines.remove(sline)
        for sline in line_list:
            self.exlines.append(sline)
        if self.exlines.closed == True:
            self.close()




import glob
import os
import pdb
import re
import shutil
import shlex
import subprocess
import sys
from urllib.parse import urlparse
from urllib.parse import ParseResult
import zipfile

from sys import platform as platform

import requests

WINDOWS_PLATFORMS = ['win32', 'win64']
LINUX_PLATFORMS = ['linux', 'linux2']

def printos(os, bits):
    """Prints the operating system and bit size"""
    print('Detected {} {} bit'.format(os, bits))

def add():
    """Installs the fea tools on windows/mac/linux"""
    osname = None
    is_64bit = sys.maxsize > 2**32
    bitsize_dict = {True: 64, False: 32}
    bitsize = bitsize_dict[is_64bit]
    if platform in LINUX_PLATFORMS:
        printos('Linux', bitsize)
        ubuntu_add()
    elif platform == "darwin":
        printos('Mac OS X', bitsize)
        mac_add()
    elif platform in WINDOWS_PLATFORMS:
        printos('Windows', bitsize)
        windows_add(bitsize)
    print('Done!')

def remove():
    """Removes the fea tools on windows/mac/linux"""
    osname = None
    is_64bit = sys.maxsize > 2**32
    bitsize_dict = {True: 64, False: 32}
    bitsize = bitsize_dict[is_64bit]
    if platform in LINUX_PLATFORMS:
        printos('Linux', bitsize)
        ubuntu_remove()
    elif platform == "darwin":
        printos('Mac OS X', bitsize)
        mac_remove()
    elif platform in WINDOWS_PLATFORMS:
        printos('Windows', bitsize)
        windows_remove(bitsize)
    print('Done!')

def ubuntu_add():
    """Adds ccx and gmsh programs on ubuntu, uses apt"""
    gmsh_installed = shutil.which('gmsh')
    if not gmsh_installed:
        print('Installing gmsh')
        command_line = "sudo apt-get install gmsh"
        subprocess.check_call(command_line, shell=True)
    else:
        print('gmsh present')
    ccx_installed = shutil.which('ccx')
    if not ccx_installed:
        print('Installing calculix (ccx)')
        command_line = "sudo apt-get install calculix-ccx"
        subprocess.check_call(command_line, shell=True)
    else:
        print('calculix (ccx) present')

def ubuntu_remove():
    """Removes ccx and gmsh programs on ubuntu, uses apt"""
    ccx_installed = shutil.which('ccx')
    if not ccx_installed:
        print('calculix (ccx) is not on your system')
    else:
        print('Removing calculix (ccx)')
        command_line = "sudo apt-get remove calculix-ccx"
        subprocess.check_call(command_line, shell=True)
    gmsh_installed = shutil.which('gmsh')
    if not gmsh_installed:
        print('gmsh is not on your system')
    else:
        print('Removing gmsh')
        command_line = "sudo apt-get remove gmsh"
        subprocess.check_call(command_line, shell=True)

def mac_add():
    """Adds ccx and gmsh programs on mac, uses brew"""
    brew_installed = shutil.which('brew')
    if not brew_installed:
        print('Installing brew')
        url = 'https://raw.githubusercontent.com/Homebrew/install/master/install'
        command_line = "/usr/bin/ruby -e \"$(curl -fsSL %s)\"" % url
        subprocess.check_call(command_line, shell=True)
    else:
        print('brew present')
    gmsh_installed = shutil.which('gmsh')
    if not gmsh_installed:
        print('Installing gmsh')
        folder_path = os.path.dirname(os.path.abspath(__file__))
        dmginstall_path = os.path.join(folder_path, 'dmginstall.sh')
        url = 'http://gmsh.info/bin/MacOSX/gmsh-3.0.5-MacOSX.dmg'
        command_line = '%s %s' % (dmginstall_path, url)
        print('command_line=%s' % command_line)
        subprocess.check_call(command_line, shell=True)
        gmsh_path = '/Applications/Gmsh.app/Contents/MacOS/gmsh'
        command_line = "ln -s %s /usr/local/bin/gmsh" % gmsh_path
        subprocess.check_call(command_line, shell=True)
    else:
        print('gmsh present')
    ccx_installed = shutil.which('ccx')
    if not ccx_installed:
        mac_add_ccx()
    else:
        print('calculix (ccx) present')

def mac_add_ccx():
    print('Installing calculix (ccx)')
    command_line = "brew install brewsci/science/calculix-ccx"
    subprocess.check_call(command_line, shell=True)
    ccx_path = find_brew_binary_location('calculix-ccx', 'ccx')
    if not ccx_path:
        raise Exception('Failed to find ccx binary')
    # link the ccx command to the ccx binary
    command_line = "ln -s %s /usr/local/bin/ccx" % ccx_path
    subprocess.check_call(command_line, shell=True)
    # check to see if we have the fortran that ccx needs and return if we do
    gcc7_to_path = "/usr/local/opt/gcc/lib/gcc/7"
    needed_fortran_path = "%s/libgfortran.4.dylib" % gcc7_to_path
    if os.path.isfile(needed_fortran_path):
        if os.path.islink(link_to_path):
            print('Linked gcc7 fortran found and used')
        else:
            print('System gcc7 fortran found and used')
        return
    # install gcc@7 with brew if we don't have it
    brew_fortran_path = ("/usr/local/Cellar/gcc@7/7.3.0/lib/gcc/7/"
                         "libgfortran.4.dylib")
    if not os.path.isfile(brew_fortran_path):
        command_line = "brew install gcc@7"
        subprocess.check_call(command_line, shell=True)
        print('Installed gcc@7 (needed by calculix)')
    # link gcc@7 from_path to to_path
    gcc7_from_path = "/usr/local/Cellar/gcc@7/7.3.0/lib/gcc/7"
    command_line = "ln -s %s %s" % (gcc7_from_path, gcc7_to_path)
    subprocess.check_call(command_line, shell=True)
    print('Finished installing calculix (ccx)')


def find_brew_binary_location(package_folder, search_string):
    """Finds the location of a binary installed by homebrew"""
    match_str = '/usr/local/Cellar/%s/**/*%s*' % (package_folder,
                                             search_string)
    paths = glob.glob(match_str, recursive=True)
    for path in paths:
        if os.access(path, os.X_OK):
            return path
    return None

def mac_remove():
    """Removes ccx and gmsh programs on mac, uses brew"""
    brew_installed = shutil.which('brew')
    if not brew_installed:
        print('Installing brew')
        url = 'https://raw.githubusercontent.com/Homebrew/install/master/install'
        command_line = "/usr/bin/ruby -e \"$(curl -fsSL %s)\"" % url
        subprocess.check_call(command_line, shell=True)
    else:
        print('brew present')
    ccx_installed = shutil.which('ccx')
    if not ccx_installed:
        print('calculix (ccx) is not on your system')
    else:
        print('Removing calculix (ccx)')
        # remove link to ccx
        command_line = "rm /usr/local/bin/ccx"
        subprocess.check_call(command_line, shell=True)
        print('Calculix (ccx) symlink was removed')
        # remove binary
        command_line = "brew uninstall calculix-ccx"
        subprocess.check_call(command_line, shell=True)
        print('Calculix (ccx) binary was removed')
        # remove gcc7 link if it exists
        gcc7_sys_path = "/usr/local/opt/gcc/lib/gcc/7"
        if os.path.islink(gcc7_sys_path):
            command_line = "rm %s" % gcc7_sys_path
            subprocess.check_call(command_line, shell=True)
            print('gcc@7 symlink was removed (it was needed by calculix)')
        # remove gcc@7 if it exists
        gcc7_brew_path = "/usr/local/Cellar/gcc@7"
        if os.path.isdir(gcc7_brew_path):
            command_line = "brew uninstall gcc@7"
            subprocess.check_call(command_line, shell=True)
            print('gcc@7 was removed (it was needed by calculix)')

    gmsh_installed = shutil.which('gmsh')
    if not gmsh_installed:
        print('gmsh is not on your system')
    else:
        print('Removing gmsh')
        command_line = "rm /usr/local/bin/gmsh"
        subprocess.check_call(command_line, shell=True)
        command_line = "rm -rf /Applications/Gmsh.app"
        subprocess.check_call(command_line, shell=True)


def windows_add(bitsize):
    """Adds ccx and gmsh programs on windows"""
    gmsh_installed = shutil.which('gmsh')
    if not gmsh_installed:
        print('Installing gmsh')
        url = 'http://gmsh.info/bin/Windows/'
        win_add_gmsh(bitsize, url, 'gmsh', '3.0.5')
    else:
        print('gmsh present')

    ccx_installed = shutil.which('ccx')
    if not ccx_installed:
        print('Installing calculix (ccx)')
        url = 'https://sourceforge.net/projects/calculixforwin/files/03.2/'
        win_add_ccx(bitsize, url, 'ccx')
    else:
        print('calculix (ccx) present')

def windows_remove(bitsize):
    """Removes ccx and gmh programs on windows"""
    gmsh_installed = shutil.which('gmsh')
    if not gmsh_installed:
        print('gmsh is not on your system')
    else:
        print('Removing gmsh')
        env_path = os.getenv('VIRTUAL_ENV', sys.exec_prefix)
        scripts_folder = '%s\Scripts\\' % env_path
        remove_like(scripts_folder, 'gmsh')
    ccx_installed = shutil.which('ccx')
    if not ccx_installed:
        print('ccx is not on your system')
    else:
        print('Removing ccx')
        env_path = os.getenv('VIRTUAL_ENV', sys.exec_prefix)
        scripts_folder = '%s\Scripts\\' % env_path
        ccx_folder = '%sccx' % scripts_folder
        add_remove_dll_links(ccx_folder, scripts_folder, add=False)
        remove_like(scripts_folder, 'ccx')

def remove_like(search_path, name):
    """
    Searches for files or folders matching name in search_path and deletes them
    """
    match_str = '%s*%s*' % (search_path, name)
    paths = glob.glob(match_str)
    for path in paths:
        print('Removing %s' % path)
        if os.path.isfile(path):
            os.unlink(path)
        elif os.path.isdir(path):
            shutil.rmtree(path)

def win_add_gmsh(bitsize, binaries_url, program_name, version_str):
    """Installs a program from an apache server file listing page"""
    # Needs the user agent header to exist for it to send back the content
    user_agent = ('Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:56.0) '
                  'Gecko/20100101 Firefox/56.0')
    headers = {'User-Agent': user_agent}

    zipfile_regex = '.*%s-%s.*' % (program_name, version_str)
    # Windows 10 64 bit, gmsh version: Gmsh 3.0.7-git-35176e26 hangs
    zipfile_name = zipfile_by_bitsize(binaries_url, headers, zipfile_regex,
                                      bitsize)

    zipfile_url = binaries_url + zipfile_name
    print('Downloading %s from %s' % (program_name, zipfile_url))
    response = requests.get(zipfile_url, stream=True, headers=headers)
    with open(zipfile_name, 'wb') as out_file:
        for chunk in response.iter_content(chunk_size=1024):
            if chunk: # filter out keep-alive new chunks
                out_file.write(chunk)
    print('Unzipping %s' % program_name)
    zip_ref = zipfile.ZipFile(zipfile_name, 'r')
    zipfile_folder_name = os.path.dirname(zip_ref.namelist()[0])
    zip_ref.extractall(None)
    zip_ref.close()
    print('Removing %s zipfile' % program_name)
    os.remove(zipfile_name)

    env_path = os.getenv('VIRTUAL_ENV', sys.exec_prefix)
    folder_to = '%s\Scripts\%s' % (env_path, zipfile_folder_name)
    exe_loc = '%s\%s.exe' % (folder_to, program_name)
    exe_link = '%s\Scripts\%s.exe' % (env_path, program_name)
    paths = [exe_link, folder_to]
    for path in  paths:
        if os.path.isfile(path):
            os.unlink(path)
        elif os.path.isdir(path):
            shutil.rmtree(path)
    print('Installing %s to %s' % (program_name, folder_to))
    shutil.move(zipfile_folder_name, folder_to)
    os.link(exe_loc, exe_link)

def zipfile_by_bitsize(binaries_url, headers, zipfile_regex, bitsize):
    """Returns the url linking to the correct zipfile"""
    # this is used by ccx and gmsh
    res = requests.get(binaries_url, headers=headers)
    html = res.text
    urls = re.findall(r'href=[\'"]?([^\'" >]+)', html)
    pattern = re.compile(zipfile_regex)
    urls = [url for url in urls if pattern.match(url)]
    urls = urls[-2:]
    url_choices = {32: urls[0], 64: urls[1]}
    if 'win32' in urls[1] or 'Windows32' in urls[1]:
        url_choices = {32: urls[1], 64: urls[0]}
    return url_choices[bitsize]

def href_from_link_text(url, headers, link_text):
    """
    Returns the url for a link with link_text description, if this
    function fails, it raises an error with tells the users a link
    which creates an issue on the pycalculix repo
    """
    response = requests.get(url, headers=headers)
    html = response.text.replace("\"", "'")
    link_text_pos = html.find(link_text)
    if link_text_pos == -1:
        issue_url = ('https://github.com/spacether/pycalculix/issues/new?title='
                     'CCX%20zip%20download%20fails%20on%20windows&body=Please%2'
                     '0update%20the%20installer.py%20get_direct_url%20function.'
                     '%20It%20no%20longer%20works')
        raise ValueError('Unable to download file because there was not a link '
                         'with text=\'%s\' on the page at url=%s\nTo fix this, '
                         'please click the \'Submit new issue\' button '
                         'here:\n%s' % (link_text, url, issue_url))
    href_pos = html[:link_text_pos].rfind('href')
    first_char = html.find("'", href_pos)+1
    last_char = html.find("'", first_char)
    return html[first_char:last_char]

def get_direct_url(url, headers):
    """Gets the zip direct download link from the project download page"""
    direct_download_url = href_from_link_text(url,
                                              headers,
                                              'Problems Downloading')
    parsed_download_url = urlparse(direct_download_url)
    if parsed_download_url.scheme not in ['http', 'https']:
        # url is relative, and is missing the scheme and netloc
        parsed_parent_url = urlparse(url)
        parsed_download_url = ParseResult(parsed_parent_url.scheme,
                                          parsed_parent_url.netloc,
                                          parsed_download_url.path,
                                          parsed_download_url.params,
                                          parsed_download_url.query,
                                          parsed_download_url.fragment)
        direct_download_url = parsed_download_url.geturl()
    direct_download_url = href_from_link_text(direct_download_url,
                                              headers,
                                              'direct link')
    return direct_download_url

def win_add_ccx(bitsize, binaries_url, program_name):
    """Installs ccx on a windows computer"""
    # Needs the user agent header to exist for it to send back the content
    user_agent = ('Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:56.0) '
                  'Gecko/20100101 Firefox/56.0')
    headers = {'User-Agent': user_agent}

    zipfile_regex = '.+CL.+win.+zip/download$'
    zipfile_webpage_url = zipfile_by_bitsize(binaries_url, headers,
                                             zipfile_regex, bitsize)

    zipfile_name = zipfile_webpage_url.split('/')[-2]
    zipfile_url = get_direct_url(zipfile_webpage_url, headers)

    print('Downloading %s from %s' % (zipfile_name, zipfile_url))
    response = requests.get(zipfile_url, stream=True, headers=headers)
    with open(zipfile_name, 'wb') as out_file:
        for chunk in response.iter_content(chunk_size=1024):
            if chunk: # filter out keep-alive new chunks
                out_file.write(chunk)
    print('Unzipping %s' % program_name)
    zip_ref = zipfile.ZipFile(zipfile_name, 'r')
    zipfile_folder_name = zip_ref.namelist()[0].split('/')[0]
    zipfile_ccx_path = zipfile_folder_name + '/bin/ccx'
    for member in zip_ref.namelist():
        if zipfile_ccx_path in member:
            zip_ref.extract(member)
    zip_ref.close()
    print('Removing %s zipfile' % program_name)
    os.remove(zipfile_name)

    folder_from = '%s\\bin\\ccx' % zipfile_folder_name
    env_path = os.getenv('VIRTUAL_ENV', sys.exec_prefix)
    scripts_path = '%s\Scripts' % env_path
    folder_to = '%s\%s' % (scripts_path, program_name)

    binary_name = 'ccx212.exe'
    exe_loc = '%s\%s' % (folder_to, binary_name)
    exe_link = '%s\%s.exe' % (scripts_path, program_name)
    paths = [exe_link, folder_to]
    for path in  paths:
        if os.path.isfile(path):
            os.unlink(path)
        elif os.path.isdir(path):
            shutil.rmtree(path)

    print('Installing %s to %s' % (program_name, folder_to))
    shutil.move(folder_from, folder_to)
    shutil.rmtree(zipfile_folder_name)
    add_remove_dll_links(folder_to, scripts_path, add=True)
    os.link(exe_loc, exe_link)

def add_remove_dll_links(folder_dlls, folder_dll_links, add=True):
    """
    Finds all dlls in folder_from and adds hard links to them in folder_to
    Or removes the hard links if add=False
    """
    match_str = '%s\*.dll' % folder_dlls
    dll_paths = glob.glob(match_str)
    for dll_path in dll_paths:
        dll_name = dll_path.split('\\')[-1]
        dll_link_path = '%s\%s' % (folder_dll_links, dll_name)
        if add:
            os.link(dll_path, dll_link_path)
        elif os.path.isfile(dll_link_path):
            os.unlink(dll_link_path)



"""This module stores the Problem class which is used to solve different types
of analysis.
"""

import os # used to see if there is a results file
import subprocess # used to launch ccx solver

from . import environment
from . import base_classes
from . import results_file

class Problem(base_classes.Idobj):
    """Makes a problem which can be analyzed with Calculix ccx.

    Args:
        - feamodel (FeaModel): the parent FeaModel
        - problem_type (str): model type, options:
            -- 'struct': structural
        - fname (str): file prefix for the problem .inp and results files
            If value is '' it will default to the project name of the FeaModel

    Attributes:
        fea (FeaModel): parent FeaModel
        __ptype (str): problem type, options:

            - 'struct': structural
        solved (bool): boolean storing whether or not the analysis was solved
        fname (str): the problem database and results file prefix
        rfile (Results_File):
            Results_File is loaded in after the model has been solved
    """
    def __init__(self, feamodel, problem_type, fname=''):
        self.fea = feamodel
        self.__ptype = problem_type
        self.solved = False
        if fname == '':
            fname = self.fea.fname
        self.fname = fname
        self.rfile = results_file.ResultsFile(self)
        base_classes.Idobj.__init__(self)
        self.fea.problems.append(self)

    @staticmethod
    def __get_ntxt(nodes):
        """Returns list of strings defining all nodes.

        Args:
            nodes (list): list of all nodes
        """
        res = []
        res.append('*NODE, NSET=nodes')
        for node in nodes:
            res.append(node.ccx())
        print('INFO: %i nodes' % len(nodes))
        return res

    def __get_etxt(self, elements):
        """Returns list of strings defining all elements.

        Args:
            elements (list): list of all elements
        """
        res = []
        ccxtypes = set([e.ccxtype for e in elements])
        eall_written = False
        eall = []
        for ccxtype in ccxtypes:
            setname = ccxtype
            if len(ccxtypes) == 1:
                setname = 'EAll'
                eall_written = True
            res.append('*ELEMENT, TYPE=%s, ELSET=%s' % (ccxtype, setname))
            elist = [e for e in elements if e.ccxtype == ccxtype]
            print('INFO: %i %s elements' % (len(elist), ccxtype))
            for element in elist:
                res.append(element.ccx())
            eall += elist
            print('INFO: %i total elements' % len(eall))
        if eall_written == False:
            tmp = self.__get_eset('EALL', eall)
            res += tmp
        return res

    @staticmethod
    def __get_ctxt(components):
        """Returns list of strings defining all components.

        Args:
            components (list): list of all components
        """
        res = []
        for comp in components:
            res += comp.ccx()
        return res

    @staticmethod
    def __get_eset(name, elements):
        """Returns list of strings defining components of elements.

        Args:
            name (str): component name
            elements (list): list of component elements
        """
        res = []
        items_per_line = 6
        res.append('*ELSET,ELSET='+name)
        grouped_els = base_classes.chunk_list(elements, items_per_line)
        for group in grouped_els:
            item_ids = [str(e.id) for e in group]
            line = ', '.join(item_ids)
            if group != grouped_els[-1]:
                line += ','
            res.append(line)
        return res

    @staticmethod
    def __fix_line(line, fstr):
        """Fixes the line and returns a fixed line.

        Args:
            line (str): line to fix
            fstr (str): format string to use
        """
        items = fstr.split(',')
        node_num_size = int(items[2][1:])
        pre_len = 3 + node_num_size
        # we are missing + prefix on some numbers, some are too long
        start_str = line[:pre_len]
        end_str = line[pre_len:]
        fields = end_str.count('E')
        low_ind = 0
        values = []
        while len(values) < fields:
            high_ind = end_str.find('E', low_ind)+5
            value = float(end_str[low_ind:high_ind])
            values.append(value)
            low_ind = high_ind
        values = ['%12.5e' % val for val in values]
        end_str = ''.join(values)
        new_str = start_str + end_str + '\n'
        return new_str

    def __fix_frd(self):
        """Fixes the frd file on win32 systems. Text formatting fixed

        On win32 Calculix, results file formatting is  not written correctly.
        Nodal results and stresses are not written as fixed length fields.
        """
        if 'win32' in environment.CCX:
            frd_file = self.fname+'.frd'
            lines = []
            try:
                with open(frd_file, "r") as infile:
                    lines = infile.readlines()
                numlines = len(lines)
                ind = 0
                fix = False
                fstr = ''
                while ind < numlines:
                    line = lines[ind]
                    if '1PSTEP' in line:
                        # we are in a results block
                        ind += 1
                        line = lines[ind]
                        format_ = int(line.split()[-1])
                        if format_ == 0:
                            fstr = "1X,I2,I5,6E12.5"
                        elif format_ == 1:
                            fstr = "1X,I2,I10,6E12.5"
                        ind += 1
                        line = lines[ind]
                        if 'DISPR' in line or 'FORCR' in line:
                            ind += 5
                        else:
                            ind += 7
                        fix = True
                        line = lines[ind]
                    if line[:3] == ' -3':
                        fix = False
                    if fix:
                        lines[ind] = self.__fix_line(line, fstr)
                    ind += 1
                with open(frd_file, "w") as outfile:
                    outfile.writelines(lines)
                print('File %s had its formatting fixed!' % frd_file)
            except IOError:
                print('Error reading .frd file!')

    def solve(self):
        """Solves the model in Calculix ccx."""
        inp = []

        # store what results we'll be outputting for each type of analysis
        out_el = {}
        out_el['struct'] = 'E,S' # strain, stress
        out_node = {}
        out_node['struct'] = 'RF,U' # reaction forces, displacement

        if self.__ptype == 'struct':
            # store nodes, elements, components
            box = {}
            box['nodes'] = []
            box['elements'] = []
            box['components'] = set()

            # store all loads in the parts in this model
            load_dict = self.fea.loads

            # store all nodes, elements, and part element sets
            box['nodes'] = self.fea.view.nodes
            box['elements'] = self.fea.view.elements

            # store all node and element components
            for time in load_dict:
                for load in load_dict[time]:
                    if load.ltype not in ['press', 'press_fluid']:
                        box['components'].add(load.comp)
            for contact in self.fea.contacts:
                box['components'].add(contact.master_comp)
                box['components'].add(contact.slave_comp)

            box['nodes'] = self.__get_ntxt(box['nodes'])
            box['elements'] = self.__get_etxt(box['elements'])
            box['components'] = self.__get_ctxt(box['components'])

            # add text definition for nodes, elelents, components
            inp += box['nodes']+['']
            inp += box['elements']+['']
            inp += box['components']+['']

            # read in all materials
            for matl in self.fea.matls:
                inp += matl.ccx()

            # write all steps and loads
            for time in sorted(load_dict.keys()):
                if time == 0.0:
                    # this is for thicknesses and materials
                    for load in load_dict[time]:
                        inp += load.ccx()
                    for surf_interaction in self.fea.surfints:
                        inp += surf_interaction.ccx()
                    for contact in self.fea.contacts:
                        inp += contact.ccx()
                else:
                    # only write times >= 1
                    inp.append('*STEP')
                    inp.append('*STATIC')

                    for load in load_dict[time]:
                        inp += load.ccx()

                    # make output frd file for cgx
                    inp.append('*EL FILE')
                    inp.append(out_el[self.__ptype])
                    inp.append('*NODE FILE')
                    inp.append(out_node[self.__ptype])

                    # make output dat file for integration point results
                    inp.append('*EL PRINT,ELSET=EALL')
                    inp.append('S')

                    # end step
                    inp.append('*END STEP')

            # write CCX inp file to the local directory
            fname = self.fname+'.inp'
            with open(fname, 'w') as outfile:
                for line in inp:
                    #print (line)
                    outfile.write(line+'\n')
            print('File: %s was written' % (fname))

            # run file
            runstr = "%s %s" % (environment.CCX, self.fname)
            print(runstr)
            subprocess.check_call(runstr, shell=True)
            print('Solving done!')

            # select the probem's parts and load the results file
            frd_file = self.fname+'.frd'
            if os.path.isfile(frd_file):
                self.solved = True
                self.__fix_frd()
                self.rfile.load()
            else:
                print('ERROR: results .frd file was not written!')



"""This module stores connector classes, like Contact.
"""
from . import base_classes

class SurfaceInteraction(base_classes.Idobj):
    """Makes a surface interaction object.

    Args:
        int_type (str): interaction type

            * 'EXPONENTIAL'
            * 'LINEAR'

        *args: following arguments

            * int_type = 'EXPONENTIAL'
                * c0, then p0 must be passed

            * int_type = 'LINEAR'
                * k must be passed

    """

    def __init__(self, int_type, *args):
        self.int_type = int_type
        self.c0 = None
        self.p0 = None
        self.k = None
        if self.int_type == 'EXPONENTIAL':
            self.c0 = args[0]
            self.p0 = args[1]
        elif self.int_type == 'LINEAR':
            self.k = args[0]
        base_classes.Idobj.__init__(self)

    @property
    def name(self):
        """SurfaceInteraction name."""
        return 'SI%i' % self.id

    def ccx(self):
        """Writes the surface interaction for the ccx solver."""
        res = ["*SURFACE INTERACTION,NAME=%s" % self.name]
        res.append("*SURFACE BEHAVIOR,PRESSURE-OVERCLOSURE=%s" % self.int_type)
        if self.int_type == 'EXPONENTIAL':
            res.append("%f,%f" % (self.c0, self.p0))
        elif self.int_type == 'LINEAR':
            res.append("%e" % self.k)
            #res.append("%e,1." % self.k)
        return res

class Contact(base_classes.Idobj):
    """Makes a contact which will be between lines which have faces on them.

    Args:
        master_comp (Component): component of master element faces
        slave_comp (Component): component of slave element faces
        surf_int (SurfaceInteraction): object which stores closure behavior
        surf_to_surf (bool): if True surface to surface is used, if False node
            to surface is used
    """

    def __init__(self, master_comp, slave_comp, surf_int, surf_to_surf=True):
        self.master_comp = master_comp
        self.slave_comp = slave_comp
        self.surf_int = surf_int
        self.surf_to_surf = surf_to_surf
        base_classes.Idobj.__init__(self)

    @property
    def name(self):
        """Contact name."""
        return 'CONT%i' % self.id

    def ccx(self):
        """Writes the contact pair for the ccx solver."""
        res = []
        end = ''
        int_name = self.surf_int.name
        m_name = self.master_comp.name
        s_name = self.slave_comp.name
        if self.surf_to_surf:
            end = ',TYPE=SURFACE TO SURFACE'
        line = "*CONTACT PAIR,INTERACTION=%s%s" % (int_name, end)
        res.append(line)
        line = "%s,%s" % (s_name, m_name)
        res.append(line)
        return res



"""This module stores the Results_File class."""

import collections
import math # used for metric number conversion
import os #need to check if results file exists
import re # used to get info from frd file
import subprocess # used to check ccx version

import matplotlib.pyplot as plt
import matplotlib.colors as colors
import matplotlib.cm as cmx
import numpy as np
from numpy.lib.polynomial import roots # need to find S1-S3
from numpy.core.function_base import linspace # need to make contours


from . import base_classes # needed for RESFIELDS
from . import environment
from . import mesh

CMAP = 'jet'

class ResultsFile(object):
    """Makes a results file.

    Args:
        problem (Problem): problem that was solved

    Attributes:
        __problem (Problem): parent problem
        __steps (list): a list of float time steps
        __results (dict): a dict storing the results data
            results[step]['node'][nnum][field] --> value
            results[step]['element'][enum]['avg'][field] --> value
            results[step]['element'][enum]['max'][field] --> value
            results[step]['element'][enum]['min'][field] --> value
            results[step]['element'][enum]['ipoints'][ipnum][field] --> value
            field = 'ux' or 'Seqv' or 'ey' etc.
        __time (float): current time we are looking at, defaults to -1
            When a file is loaded in, the first time step is loaded.
    """

    def __init__(self, problem):
        self.__problem = problem
        self.__steps = [] # this stores a list of time steps
        self.__results = {} # stores results, nested dicts
        self.__time = -1
        if self.__problem.solved:
            self.load()

    @property
    def steps(self):
        """Returns a list of loaded time steps.

        Note: this is read only, you can not assign a value to it.
        """
        return self.__steps

    @property
    def time(self):
        """Returns the current time (float) in the results file.

        Note: this is read only, you can not assign a value to it.
        """
        return self.__time

    @staticmethod
    def __metric_num(number, sig_figs=3, sci=False):
        """Returns string of number, with only 10**3 suffixes.

        If 0 <= number < 1000 no suffix is added.
        This is useful for quantities usually given in metric: stress, displacement.

        Args:
            number (float or int): the number we want converted to __metric_number
            sig_figs (int): number of significant figures to use, right of decimal
            sci (bool): True means use scientific formatting, False use metric
        """
        if sci:
            format_str = "%.{}e".format(sig_figs)
            my_str = format_str % number
        else:
            format_str = "%.{}f".format(sig_figs)
            my_str = format_str % number
            if number != 0:
                # get the scientific exponent
                exp = math.floor(math.log10(abs(number)))
                metric_exp = exp - (exp % 3)
                new_float = number/(10**metric_exp)
                if metric_exp != 0:
                    format_str = "%.{}fe%i".format(sig_figs)
                    my_str = format_str % (new_float, metric_exp)
        return my_str

    def load(self):
        """Loads the results file with problem.fname prefix."""
        self.__read_frd() # read nodal results
        self.__read_dat() # read element integration pt results

    def nplot(self, field, fname='', display=True, levels=21, gradient=False,
              gmult=1.0, max_val=None, min_val=None, title=''):
        """Plots nodal results.

        Args:
            field (str): results item to plot, examples: 'ux', 'ey', 'Seqv'
            fname (str): prefix of png file name, if writing an image
            display (bool): True = interactively show the plot
            levels (int): number of levels to use in the colorbar
            gradient (bool): True = results plotted with gradient
                False = results plotted with filled areas
            gmult (int): geometric multiplier on displacement of nodes
                displayed_node_loc = model_node_loc + gmult*node_displacement
            max_val (float or None): max value in the colorbar

                - None: max from selected data used
                - float: use the passed float
            min_val (float): min value in the colorbar

                - None: min from selected data used
                - float: use the passed float
            title (str): third line in the plot title
        """
        # store the selected nodes and elements
        sel = {}
        sel['nodes'] = self.__problem.fea.view.nodes
        sel['elements'] = self.__problem.fea.view.elements
        sel['faces'] = self.__problem.fea.view.faces

        # sort nodes low to high so index is correct
        # we have index to id below so showing subsets works
        sel['nodes'] = list(sel['nodes'])
        sel['nodes'] = sorted(sel['nodes'], key=lambda k: k.id)

        # store results at nodes
        axials = []
        radials = []
        zvals = []
        id_to_ind = {}
        for node in sel['nodes']:
            id_to_ind[node.id] = len(axials)
            axi = node.y + gmult*self.__results[self.__time]['node'][node.id]['uy']
            rad = node.x + gmult*self.__results[self.__time]['node'][node.id]['ux']
            axials.append(axi)
            radials.append(rad)
            zvals.append(self.__results[self.__time]['node'][node.id][field])

        # make a list of triangles, given by indices, looping anticlockwise
        triangles = []
        mylist = []
        if len(sel['elements']) > 0:
            mylist = sel['elements']
        elif len(sel['faces']) > 0:
            mylist = sel['faces']
        for element in mylist:
            tris = element.get_tris() # list of triangle nodes
            for tri in tris:
                for ind, nid in enumerate(tri):
                    tri[ind] = id_to_ind[nid]     # convert id to index
            triangles += tris

        # check to see if selected nodes and elements are
        # in the parent model's nodes and elements
        fig = plt.figure()
        ax_ = fig.add_subplot(111)

        # need to set tick list here
        vmin = min(zvals)
        vmax = max(zvals)
        stop_plot = False
        if max_val != None and min_val == None:
            if max_val < vmin:
                stop_plot = True
                print('Error:')
                print(' Only max was passed but it is < the data min!')
                print(' Pass a max_val that is > the data min of %f' % vmin)
            else:
                vmax = max_val
        elif min_val != None and max_val == None:
            if min_val > vmax:
                stop_plot = True
                print('Error:')
                print(' Only min was passed but it is > the data max!')
                print(' Pass a min_val that is < the data max of %f' % vmax)
            else:
                vmin = min_val
        elif max_val != None and min_val != None:
            if max_val < min_val:
                stop_plot = True
                print('Error:')
                print(' Min and max passed, but min > max!')
                print(' Pass a min_val that is < max_val')
            else:
                vmax = max_val
                vmin = min_val
        # exit if stop plot flag is on
        if stop_plot:
            return None

        tick_list = [vmin]
        if vmax != vmin:
            # we have a range of values we're plotting
            tick_list = linspace(vmin, vmax, levels+1)

        # plot using a gradient(shaded) or levels
        # code required for the colorbar, needs to go before plotting for colormap
        cnorm = colors.Normalize(vmin=vmin, vmax=vmax)
        cmap = colors.ListedColormap(['b', 'b']) # default to plot one val
        if vmax != vmin:
            # we have a range of values we're plotting
            if gradient:
                cmap = plt.get_cmap(CMAP)
            else:
                cmap = plt.get_cmap('jet', levels)
        cmap.set_under('0.3', 0.8)
        cmap.set_over('0.7', 0.8)
        if gradient or len(tick_list) == 1:
            # This one is shaded
            plt.tripcolor(axials, radials, triangles, zvals, shading='gouraud',
                          cmap=cmap, norm=cnorm)
        else:
            # this one is not shaded
            plt.tricontourf(axials, radials, triangles, zvals, levels=tick_list,
                            cmap=cmap, norm=cnorm, extend='both')

        scalarmap = cmx.ScalarMappable(norm=cnorm, cmap=cmap)
        scalarmap.set_array([])
        cbar = plt.colorbar(scalarmap, orientation='vertical', ticks=tick_list)

        scibool = False
        if field[0] == 'e':
            # strain plotting, use scientific numbering
            scibool = True
        met_max = self.__metric_num(max(zvals), sci=scibool)
        met_min = self.__metric_num(min(zvals), sci=scibool)
        label = 'Max: %s\nMin: %s' % (met_max, met_min)
        tick_list = [self.__metric_num(tick, sci=scibool) for tick in tick_list]
        cbar.ax.set_yticklabels(tick_list)
        cbar.ax.set_xlabel(label, labelpad=10, x=0, ha='left')
        cbar.ax.xaxis.set_label_position('top')

        # set the horizontal and vertical axes
        base_classes.plot_set_bounds(plt, axials, radials)

        # set units
        alist = self.__problem.fea.get_units(field, 'dist', 'time')
        [f_unit, d_unit, t_unit] = alist

        # set plot axes
        plot_title = ('Node %s%s\nTime=%f%s' %
                      (field, f_unit, self.__time, t_unit))
        if title != '':
            plot_title += '\n%s' % title
        plt.title(plot_title)
        plt.xlabel('axial, y'+d_unit)
        plt.ylabel('radial, x'+d_unit)
        ax_.set_aspect('equal')
        if gmult != 1:
            ax_.xaxis.set_ticklabels([])
            ax_.yaxis.set_ticklabels([])
        base_classes.plot_finish(plt, fname, display)

    def eplot(self, field, fname='', display=True, levels=21,
              gmult=1.0, mode='avg', max_val=None, min_val=None, title=''):
        """Plots element results.

        Args:
            field (str): results item to plot. Only stresses supported.
                Examples: 'Sx', 'Sxy', 'S1', 'Seqv' etc.
            fname (str): prefix of png file name, if writing an image
            display (bool): True = interactively show the plot
            levels (int): number of levels to use in the colorbar
            gmult (int): geometric multiplier on displacement of nodes
                displayed_node_loc = model_node_loc + gmult*node_displacement
            mode (str): the type of element result to plot

                - 'avg': integration points averaged to avg element result
                - 'max': max value of field in the integration points plotted
                - 'min': min value of field in the integration points plotted
            max_val (float or None): max value in the colorbar

                - None: max from selected data used
                - float: use the passed float
            min_val (float): min value in the colorbar

                - None: min from selected data used
                - float: use the passed float
            title (str): third line in the plot title
        """
        # store the selected nodes and elements
        sel = {}
        sel['nodes'] = self.__problem.fea.view.nodes
        sel['elements'] = self.__problem.fea.view.elements
        sel['faces'] = self.__problem.fea.view.faces

        # sort nodes low to high so index is correct
        # we have index to id below so showing subsets works
        sel['nodes'] = list(sel['nodes'])
        sel['nodes'] = sorted(sel['nodes'], key=lambda k: k.id)

        # store results at nodes
        axials = []
        radials = []
        zvals = []
        id_to_ind = {}
        for node in sel['nodes']:
            id_to_ind[node.id] = len(axials)
            axi = node.y + gmult*self.__results[self.__time]['node'][node.id]['uy']
            rad = node.x + gmult*self.__results[self.__time]['node'][node.id]['ux']
            axials.append(axi)
            radials.append(rad)

        # make a list of triangles, given by indices, looping anticlockwise
        triangles = []
        mylist = []
        if len(sel['elements']) > 0:
            mylist = sel['elements']
        elif len(sel['faces']) > 0:
            mylist = sel['faces']
        for ele in mylist:
            val = self.__results[self.__time]['element'][ele.id][mode][field]
            tris = ele.get_tris() # list of triangle nodes defined by node id
            for tri in tris:
                zvals.append(val)
                for ind, nid in enumerate(tri):
                    tri[ind] = id_to_ind[nid]     # convert id to index
            triangles += tris

        # check to see if selected nodes and elements are
        # in the parent model's nodes and elements

        fig = plt.figure()
        ax_ = fig.add_subplot(111)

        # need to set tick list here
        vmin = min(zvals)
        vmax = max(zvals)
        stop_plot = False
        if max_val != None and min_val == None:
            if max_val < vmin:
                stop_plot = True
                print('Error:')
                print(' Only max was passed but it is < the data min!')
                print(' Pass a max_val that is > the data min of %f' % vmin)
            else:
                vmax = max_val
        elif min_val != None and max_val == None:
            if min_val > vmax:
                stop_plot = True
                print('Error:')
                print(' Only min was passed but it is > the data max!')
                print(' Pass a min_val that is < the data max of %f' % vmax)
            else:
                vmin = min_val
        elif max_val != None and min_val != None:
            if max_val < min_val:
                stop_plot = True
                print('Error:')
                print(' Min and max passed, but min > max!')
                print(' Pass a min_val that is < max_val')
            else:
                vmax = max_val
                vmin = min_val
        # exit if stop plot flag is on
        if stop_plot:
            return None

        tick_list = [vmin]
        if vmax != vmin:
            # we have a range of values we're plotting
            tick_list = linspace(vmin, vmax, levels+1)

        # code required for the colorbar, needs to go before plotting for cmap
        cnorm = colors.Normalize(vmin=vmin, vmax=vmax)
        cmap = colors.ListedColormap(['b', 'b']) # default to plot one val
        if vmax != vmin:
            # we have a range of values we're plotting
            cmap = plt.get_cmap(CMAP, levels)
        cmap.set_under('0.3', 0.8)
        cmap.set_over('0.7', 0.8)

        # plot using levels
        plt.tripcolor(axials, radials, triangles, zvals,
                      shading='flat', cmap=cmap, norm=cnorm)

        scalarmap = cmx.ScalarMappable(norm=cnorm, cmap=cmap)
        scalarmap.set_array([])
        cbar = plt.colorbar(scalarmap, orientation='vertical', ticks=tick_list)
        scibool = False
        if field[0] == 'e':
            # strain plotting, use scientific numbering
            scibool = True
        met_max = self.__metric_num(max(zvals), sci=scibool)
        met_min = self.__metric_num(min(zvals), sci=scibool)
        label = 'Max: %s\nMin: %s' % (met_max, met_min)
        tick_list = [self.__metric_num(tick, sci=scibool) for tick in tick_list]
        cbar.ax.set_yticklabels(tick_list)
        cbar.ax.set_xlabel(label, labelpad=10, x=0, ha='left')
        cbar.ax.xaxis.set_label_position('top')

        # set the horizontal and vertical axes
        base_classes.plot_set_bounds(plt, axials, radials)

        # set units
        alist = self.__problem.fea.get_units(field, 'dist', 'time')
        [f_unit, d_unit, t_unit] = alist

        # set plot axes
        plot_title = ('Element %s %s%s\nTime=%f%s' %
                      (mode, field, f_unit, self.__time, t_unit))
        if title != '':
            plot_title += '\n%s' % title
        plt.title(plot_title)
        plt.xlabel('axial, y'+d_unit)
        plt.ylabel('radial, x'+d_unit)
        ax_.set_aspect('equal')
        if gmult != 1:
            ax_.xaxis.set_ticklabels([])
            ax_.yaxis.set_ticklabels([])
        base_classes.plot_finish(plt, fname, display)

    def set_time(self, time):
        """Sets the time point we're looking at in the results file.

        Args:
            time (float): time we are setting
        """
        if time in self.steps:
            self.__time = time
            print('Results file time set to: %f' % (self.__time))
        else:
            print('Time %f is not in the loaded times. Valid times are:')
            print(self.steps)


    def plot_gradient(self, start_point, end_point, field, fname='', display=True, title='', max_val=None, min_val=None, curve_fitting=True, n_poly=3, n_subpoints=500, legend=True):
        """Create diagram with data projected onto line on the undeformed geometry.

        Args:
            start_point [float, float]: starting point of line. [x, y]
            end_point [float, float]: end point of line. Example: [x, y]
            field (str): results item to plot, examples: 'ux', 'ey', 'Seqv'
            fname (str): prefix of png file name, if writing an image
            display (bool): True = interactively show the plot
            title (str): third line in the plot title
            max_val (float or None): max value in the y-axis
                - None: max from selected data used
                - float: use the passed float
            min_val (float or None): min value in the y-axis
                - None: min from selected data used
                - float: use the passed float
            curve_fitting (bool): True = a curve is fitted to the gradient
            n_poly (int): numbers of polygons for fitting
            n_subpoints (int): numbers of points the line is subdivided into
            legend (bool): True = legend with fitted equation is shown
        """

        # store the selected nodes and elements
        sel = {}
        sel['nodes'] = self.__problem.fea.view.nodes

        # sort nodes low to high so index is correct
        # we have index to id below so showing subsets works
        sel['nodes'] = list(sel['nodes'])
        sel['nodes'] = sorted(sel['nodes'], key=lambda k: k.id)

        # store results at nodes
        node_position = np.zeros((len(sel['nodes']),2))
        field_values = np.zeros(len(sel['nodes']))

        for idx, node in enumerate(sel['nodes']):

            node_position[idx] = [node.x, node.y]
            field_values[idx] = self.__results[self.__time]['node'][node.id][field]


        #create subpoints on line
        subpoints = np.zeros((n_subpoints, 3))  #[x, y, line position]

        subpoints[:,0] = np.linspace(start_point[0], end_point[0], n_subpoints)
        subpoints[:,1] = np.linspace(start_point[1], end_point[1], n_subpoints)
        subpoints[:,2] = np.arange(n_subpoints) / n_subpoints * np.sqrt(np.sum( (np.array(start_point) - np.array(end_point))**2))

        #calculate weighted field value for every subpoint
        wfield = np.zeros(n_subpoints)

        for idx in range(n_subpoints):

            #calculate inverse of distance from nodes to subpoints
            dist = np.sqrt(np.sum((node_position-subpoints[idx,0:2])**2,axis=1))

            #calculte weighted field value
            #dist[dist < 1E-10] = 1E-10
            #inv_dist = 1. / dist**3
            #wfield[idx] = np.average(field_values, weights=inv_dist)

            #use nearest value
            wfield[idx] = field_values[min(range(len(dist)),key=dist.__getitem__)]


        #plot diagram

        fig = plt.figure(figsize=(10,6))
        ax_ = fig.add_subplot(111)

        plt.plot(subpoints[:,2], wfield, '-r', linewidth=2.5, label=field)

        if curve_fitting==True:
            #execute curve fitting if needed
            poly = np.polyfit(subpoints[:,2], wfield, n_poly)

            #string for equation of fitted function
            funcstring = [str(np.round(poly[i]))+u'*x^'+str(np.arange(n_poly,0,-1)[i]) for i in range(n_poly)]
            funcstring.append(str(np.round(poly[-1])))
            funcstring = '+'.join(funcstring)

            func = np.poly1d(poly)

            plt.plot(subpoints[:,2], func(subpoints[:,2]), '--k', linewidth=1.5, label=funcstring)


        # set units
        alist = self.__problem.fea.get_units(field, 'dist', 'time')
        [f_unit, d_unit, t_unit] = alist

        # set plot axes
        plot_title = ('Gradient %s%s\nTime=%f%s' %(field, f_unit, self.__time, t_unit))
        if title != '':
            plot_title += '\n%s' % title
        plt.title(plot_title)
        plt.xlabel('path position'+d_unit)
        plt.ylabel(field + ' ' +f_unit)

        #show legend if needed
        if legend == True:
            plt.legend()

        #set limits on y-axis
        if min_val!=None:
            plt.gca().set_ylim(bottom=min_val)
        if max_val!=None:
            plt.gca().set_ylim(top=max_val)

        plt.grid()
        base_classes.plot_finish(plt, fname, display)

    def get_relative_gradient(self, start_point, end_point, field, n_poly=3, n_subpoints=500):
        """Calculte relative stress gradient (gradient/start_value)

        Args:
            start_point [(float), (float)]: starting point of line. [x, y]
            end_point [(float), (float)]: end point of line. Example: [x, y]
            field (str): results item to plot, examples: 'ux', 'ey', 'Seqv'

        Kargs:
            n_poly (int): numbers of polygons for fitting, min=2
            n_subpoints (int): numbers of points the line is subdivided into
        """

        # store the selected nodes and elements
        sel = {}
        sel['nodes'] = self.__problem.fea.view.nodes

        # sort nodes low to high so index is correct
        # we have index to id below so showing subsets works
        sel['nodes'] = list(sel['nodes'])
        sel['nodes'] = sorted(sel['nodes'], key=lambda k: k.id)

        # store results at nodes
        node_position = np.zeros((len(sel['nodes']),2))
        field_values = np.zeros(len(sel['nodes']))

        for idx, node in enumerate(sel['nodes']):

            node_position[idx] = [node.x, node.y]
            field_values[idx] = self.__results[self.__time]['node'][node.id][field]


        #create subpoints on line
        subpoints = np.zeros((n_subpoints, 3))  #[x, y, line position]

        subpoints[:,0] = np.linspace(start_point[0], end_point[0], n_subpoints)
        subpoints[:,1] = np.linspace(start_point[1], end_point[1], n_subpoints)
        subpoints[:,2] = np.arange(n_subpoints) / n_subpoints * np.sqrt(np.sum( (np.array(start_point) - np.array(end_point))**2))

        #calculate weighted field value for every subpoint
        wfield = np.zeros(n_subpoints)

        for idx in range(n_subpoints):

            #calculate inverse of distance from nodes to subpoints
            dist = np.sqrt(np.sum((node_position-subpoints[idx,0:2])**2,axis=1))

            #use nearest value
            wfield[idx] = field_values[min(range(len(dist)),key=dist.__getitem__)]


        #curve fitting
        poly = np.polyfit(subpoints[:,2], wfield, n_poly)

        rel_grad = abs(poly[-2])/abs(poly[-1])

        return rel_grad





    @staticmethod
    def __utot(vals):
        """Returns the total displacement distance, given [dx,dy,dz].

        Args:
            vals (list): [dx, dy, dz] list of displacements in x, y, and z axes

        Returns:
            res (float): displacement
        """
        # computes sum of the squares
        res = [a**2 for a in vals]
        res = (sum(res))**0.5
        return res

    @staticmethod
    def __seqv(vals):
        """Returns the Von Mises stress, which will be stored as 'Seqv'.

        Args:
            vals (list): list of six stresses [s11,s22,s33,s12,s13,s23]

        Returns:
            res (float): Von Mises stress
        """
        [s11, s22, s33, s12, s13, s23] = vals
        aval = s11 - s22
        bval = s22 - s33
        cval = s33 - s11
        dval = s12**2 + s23**2 +s13**2
        res = (0.5*(aval**2 + bval**2 + cval**2 +6*dval))**0.5
        return res

    @staticmethod
    def __principals(vals):
        """Returns principal stresses [S1,S2,S3].

        Args:
            vals (list): six stresses [s11,s22,s33,s12,s13,s23]

        Returns:
            res (list): principal stresses [S1,S2,S3] stresses are high-to-low
        """
        # calculates and returns principal stresses, S1, S2, S3
        [s11, s22, s33, s12, s13, s23] = vals
        aval = 1
        bval = (s11 + s22 + s33)*-1.0
        cval = (s11*s22 + s11*s33 + s22*s33 - s12**2 - s13**2 - s23**2)
        dval = (s11*s22*s33 + 2*s12*s13*s23 - s11*(s23**2) - s22*(s13**2)
                - s33*(s12**2))*-1.0
        res = list(roots([aval, bval, cval, dval]))
        res = sorted(res, reverse=True)
        return res

    def __get_data_dict(self, time, type_str):
        """Returns the data dict at the correct time for element or node.

        Args:
            time (float): None or the time we want, if None use current time
            type_str: 'element' or 'node'

        Returns:
            res (dict or None): dictionary with field values in it
                None if the time was invalid
        """
        res = self.__results[self.__time][type_str]
        if time != None:
            if time not in self.steps:
                print('Error: passed time is not in steps!')
                print(' Pass a time in the steps:')
                print(self.steps)
                return None
            else:
                res = self.__results[time][type_str]
        return res

    def get_nmax(self, field, time=None):
        """Returns the max value of node results field in selected nodes.

        Reports results for the current time.

        Args:
            field (str): results field, for example 'ux', 'ey', 'S1', 'Seqv'
            time (None or float): the time to query

                - None: uses the current time
                - float: uses the passed float time
        Returns:
            res (float): max value
        """
        nodes = self.__problem.fea.view.nodes
        node_ids = [node.id for node in nodes]
        data_dict = self.__get_data_dict(time, 'node')
        if data_dict == None:
            return None
        ndicts = [data_dict[nid] for nid in node_ids]
        res = [ndict[field] for ndict in ndicts]
        res = max(res)
        return res

    def get_nmin(self, field, time=None):
        """Returns the min value of node results field in selected nodes.

        Reports results for the current time.

        Args:
            field (str): results field, for example 'ux', 'ey', 'S1', 'Seqv'
            time (None or float): the time to query

                - None: uses the current time
                - float: uses the passed float time
        Returns:
            res (float): min value
        """
        nodes = self.__problem.fea.view.nodes
        node_ids = [node.id for node in nodes]
        data_dict = self.__get_data_dict(time, 'node')
        if data_dict == None:
            return None
        ndicts = [data_dict[nid] for nid in node_ids]
        res = [ndict[field] for ndict in ndicts]
        res = min(res)
        return res

    def get_nval(self, node, field, time=None):
        """Returns the field result value under node.

        Result will be returned whether or not passed node is selected.

        Args:
            node (str or Node): node we are asking about
            field (str): the results item we want: 'ux', 'Sy', 'Seqv', 'fx'
            time (None or float): the time to query

                - None: uses the current time
                - float: uses the passed float time
        Returns:
            res (float or None): float value if field exists, None otherwise
        """
        items = self.__problem.fea.get_item(node)
        if len(items) == 1:
            if isinstance(items[0], mesh.Node):
                nnum = items[0].id
                data_dict = self.__get_data_dict(time, 'node')
                if data_dict == None:
                    return None
                ndict = data_dict[nnum]
                if field in ndict:
                    res = ndict[field]
                    return res
                else:
                    print('Passed field is not in the results!')
                    return None
            else:
                print('You did not pass in a node!')
                print('A single node or string node name must be passed in!')
                return None
        else:
            print('A single node or string node name must be passed in!')
            return None

    def get_fsum(self, item):
        """Returns the force sum on nodes under a given point or line.

        Reports results for the current time.

        Args:
            item (Point or SignLine): item that has reaction forces on its nodes

        Returns:
            list: [fx, fy, fz] reaction forces in each axis, force units
        """
        (fxx, fyy, fzz) = ([], [], [])
        nodes = item.nodes
        nodes = [n.id for n in nodes]
        for node in nodes:
            f_x = self.__results[self.__time]['node'][node]['fx']
            f_y = self.__results[self.__time]['node'][node]['fy']
            f_z = self.__results[self.__time]['node'][node]['fz']
            if f_x != 0 or f_y != 0 or f_z != 0:
                fxx.append(f_x)
                fyy.append(f_y)
                fzz.append(f_z)
        fxx = sum(fxx)
        fyy = sum(fyy)
        fzz = sum(fzz)
        return [fxx, fyy, fzz]

    def get_emax(self, field, time=None, mode='avg'):
        """Returns the max results field value of selected elements at curent time.

        Args:
            field (str): results field, stresses supported 'S1', 'Sx', etc.
            time (None or float): the time to query

                - None: uses the current time
                - float: uses the passed float time
            mode (str): type of element result to give back

                - 'max': for each element only use the max value of field over
                    all of its integration points
                - 'min': for each element only use the min value of field over
                    all of its integration points
                - 'avg': for each element only use an average of all integration
                    points in the eleemnt. Principal streses and Seqv are
                    calculated after averaging 6 stress components.
        Returns:
            res (float): max value
        """
        res = []
        elements = self.__problem.fea.view.elements
        data_dict = self.__get_data_dict(time, 'element')
        if data_dict == None:
            return None
        for element in elements:
            enum = element.id
            edict = data_dict[enum][mode]
            res.append(edict[field])
        res = max(res)
        return res

    def get_emin(self, field, time=None, mode='avg'):
        """Returns the min results field value of selected elements at curent time.

        Args:
            field (str): results field, stresses supported 'S1', 'Sx', etc.
            time (None or float): the time to query

                - None: uses the current time
                - float: uses the passed float time
            mode (str): type of element result to give back

                - 'max': for each element only use the max value of field over
                    all of its integration points
                - 'min': for each element only use the min value of field over
                    all of its integration points
                - 'avg': for each element only use an average of all integration
                    points in the eleemnt. Principal streses and Seqv are
                    calculated after averaging 6 stress components.
        Returns:
            res (float): min value
        """
        res = []
        elements = self.__problem.fea.view.elements
        data_dict = self.__get_data_dict(time, 'element')
        if data_dict == None:
            return None
        for element in elements:
            enum = element.id
            edict = data_dict[enum][mode]
            res.append(edict[field])
        res = min(res)
        return res

    def get_eval(self, element, field, time=None, mode='avg'):
        """Returns the field result value under element.

        Result will be returned whether or not passed element is selected.

        Args:
            element (str or Element): element we are asking about
            field (str): the results item we want: 'Sy', 'Seqv'
            mode (str): the type of element result to get

                - 'avg': integration points averaged to avg element result
                - 'max': max value of field in the integration points plotted
                - 'min': min value of field in the integration points plotted
        Returns:
            res (float or None): float value if field exists, None otherwise
        """
        items = self.__problem.fea.get_item(element)
        if len(items) == 1:
            if isinstance(items[0], mesh.Element):
                enum = items[0].id
                data_dict = self.__get_data_dict(time, 'element')
                if data_dict == None:
                    return None
                edict = data_dict[enum][mode]
                if field in edict:
                    res = edict[field]
                    return res
                else:
                    print('Passed field is not in the results!')
                    return None
            else:
                print('You did not pass in a element!')
                print('A single element or string element name must be given!')
                return None
        else:
            print('A single element or string element name must be given!')
            return None

    @staticmethod
    def __get_vals(fstr, line):
        """Returns a list of typed items based on an input format string.

        Args:
            fst (str): C format string, commas separate fields
            line (str): line string to parse

        Returns:
            res (list): list of typed items extracted from the line
        """
        res = []
        fstr = fstr.split(',')
        thestr = str(line)
        for item in fstr:
            if item[0] == "'":
                # strip off the char quaotes
                item = item[1:-1]
                # this is a string entry, grab the val out of the line
                ind = len(item)
                fwd = thestr[:ind]
                thestr = thestr[ind:]
                res.append(fwd)
            else:
                # format is: 1X, A66, 5E12.5, I12
                # 1X is number of spaces
                (mult, ctype) = (1, None)
                m_pat = re.compile(r'^\d+') # find multiplier
                c_pat = re.compile(r'[XIEA]') # find character
                if m_pat.findall(item) != []:
                    mult = int(m_pat.findall(item)[0])
                ctype = c_pat.findall(item)[0]
                if ctype == 'X':
                    # we are dealing with spaces, just reduce the line size
                    thestr = thestr[mult:]
                elif ctype == 'A':
                    # character string only, add it to results
                    fwd = thestr[:mult].strip()
                    thestr = thestr[mult:]
                    res.append(fwd)
                else:
                    # IE, split line into m pieces
                    w_pat = re.compile(r'[IE](\d+)') # find the num after char
                    width = int(w_pat.findall(item)[0])
                    while mult > 0:
                        # only add items if we have enough line to look at
                        if width <= len(thestr):
                            substr = thestr[:width]
                            thestr = thestr[width:]
                            substr = substr.strip() # remove space padding
                            if ctype == 'I':
                                substr = int(substr)
                            elif ctype == 'E':
                                substr = float(substr)
                            res.append(substr)
                        mult -= 1
        return res

    @staticmethod
    def __get_first_dataline(infile):
        """
        Reads infile until a line with data is found, then returns it
        A line that starts with ' -1' has data
        """
        while True:
            line = infile.readline()
            if line[:3] == ' -1':
                return line

    def __store_time(self, time):
        """Stores the passed time in the results steps"""
        if time not in self.__steps:
            self.__steps.append(time)
        if time not in self.__results:
            new_dict = {'node': collections.defaultdict(dict),
                        'element': collections.defaultdict(dict)}
            self.__results[time] = new_dict

    def __modearr_estrsresults(self, infile, line):
        """Returns an array of line, mode, rfstr, time"""
        words = line.strip().split()
        # add time if not present
        time = float(words[-1])
        self.__store_time(time)
        # set mode
        rfstr = "I10,2X,I2,6E14.2"
        mode = 'stress'
        infile.readline()
        line = infile.readline()
        return [line, mode, rfstr, time]

    def __modearr_nresults(self, infile):
        """Returns an array of line, mode, rfstr, time"""
        line = infile.readline()
        fstr = "1X,' 100','C',6A1,E12.5,I12,20A1,I2,I5,10A1,I2"
        tmp = self.__get_vals(fstr, line)
        #[key, code, setname, value, numnod, text, ictype, numstp, analys, format_]
        time, format_ = tmp[3], tmp[9]

        # set results format to short, long or binary
        # only short and long are parsed so far
        if format_ == 0:
            rfstr = "1X,I2,I5,6E12.5"
        elif format_ == 1:
            rfstr = "1X,I2,I10,6E12.5"
        elif format_ == 2:
            # binary
            pass

        # set the time
        self.__store_time(time)

        # get the name to determine if stress or displ
        line = infile.readline()
        fstr = "1X,I2,2X,8A1,2I5"
        # [key, name, ncomps, irtype]
        name = self.__get_vals(fstr, line)[1]

        mode_by_name = {'DISP': 'displ',
                        'STRESS': 'stress',
                        'TOSTRAIN': 'strain',
                        'FORC': 'force'}
        mode = mode_by_name[name]
        print('Reading '+mode+' storing: '+
              ','.join(base_classes.RESFIELDS[mode]))

        line = self.__get_first_dataline(infile)
        return [line, mode, rfstr, time]

    def _save_node_displ(self, line, rfstr, time, mode='displ'):
        """Saves node displacement"""
        node, ux_, uy_, uz_ = self.__get_vals(rfstr, line)[1:]
        labs = base_classes.RESFIELDS[mode]
        vals = [ux_, uy_, uz_]
        utot = self.__utot(vals)
        vals.append(utot)
        adict = self.__results[time]['node'][node]
        for (label, val) in zip(labs, vals):
            adict[label] = val

    def _save_node_stress(self, line, rfstr, time, mode='stress'):
        """Saves node stress"""
        tmp = self.__get_vals(rfstr, line)
        # [key, node, sx, sy, sz, sxy, syz, szx]
        node, sxx, syy, szz, sxy, syz, szx = tmp[1:]
        labs = base_classes.RESFIELDS[mode]
        vals = [sxx, syy, szz, sxy, syz, szx]
        seqv = self.__seqv(vals)
        s_1, s_2, s_3 = self.__principals(vals)
        vals.append(seqv)
        vals += [s_1, s_2, s_3]
        adict = self.__results[time]['node'][node]
        for (label, val) in zip(labs, vals):
            adict[label] = val

    def _save_node_strain(self, line, rfstr, time, mode='strain'):
        """Saves node strain"""
        tmp = self.__get_vals(rfstr, line)
        # [key, node, ex, ey, ez, exy, eyz, ezx]
        node, exx, eyy, ezz, exy, eyz, ezx = tmp[1:]
        labs = base_classes.RESFIELDS[mode]
        vals = [exx, eyy, ezz, exy, eyz, ezx]
        eeqv = self.__seqv(vals)
        e_1, e_2, e_3 = self.__principals(vals)
        vals.append(eeqv)
        vals += [e_1, e_2, e_3]
        adict = self.__results[time]['node'][node]
        for (label, val) in zip(labs, vals):
            adict[label] = val

    def _save_node_force(self, line, rfstr, time, mode='force'):
        """Saves node force"""
        # [key, node, fx, fy, fz]
        node, f_x, f_y, f_z = self.__get_vals(rfstr, line)[1:]
        labs = base_classes.RESFIELDS[mode]
        vals = [f_x, f_y, f_z]
        adict = self.__results[time]['node'][node]
        for (label, val) in zip(labs, vals):
            adict[label] = val

    def _save_ele_stress(self, line, rfstr, time,
                                  mode='stress'):
        """Saves element integration point stresses"""
        labels = ['Sx', 'Sy', 'Sz', 'Sxy', 'Sxz', 'Syz']
        vals = self.__get_vals(rfstr, line)
        # element_number, integration_pt_number
        enum, ipnum = vals[0], vals[1]
        stress_vals = vals[2:]

        adict = {}
        for (label, val) in zip(labels, stress_vals):
            adict[label] = val
        if enum not in self.__results[time]['element']:
            start_val = {'ipoints': {},
                         'avg': {},
                         'min': {},
                         'max': {}}
            self.__results[time]['element'][enum] = start_val
        # each line is an integration point result
        self.__results[time]['element'][enum]['ipoints'][ipnum] = adict

    def check_ccx_version(self, timeout=1):
        """Raises an exception of the calculix ccx version is too old"""
        runstr = "%s -version" % (environment.CCX)
        try:
            output_str = subprocess.check_output(runstr,
                                                 timeout=timeout,
                                                 shell=True)
        except (subprocess.CalledProcessError, subprocess.TimeoutExpired) as ex:
            output_str = ex.output
        output_str = str(output_str, 'utf-8')
        matches = re.findall(r'\d+\.\d+', output_str)
        version_number = matches[-1]
        print('Using Calculix ccx version=%s '
              '(trailing characters like the p in 2.8p are omitted)'
              % version_number)
        major_version, minor_version = [int(f) for f
                                        in version_number.split('.')]
        if major_version <= 2 and minor_version <= 8:
            raise Exception('Your version of calculix ccx is too old! '
                            'Please update it to version >=2.8 with '
                            'the command:\npycalculix-add-feaprograms')
        version = float(output_str.strip().split()[-1])
        # extract version with regex

    def __read_frd(self):
        """
        Reads a ccx frd results file which contains nodal results.
        The file format is desribed in the cgx docs
        """
        fname = self.__problem.fname+'.frd'
        if not os.path.isfile(fname):
            print("Error: %s file not found" % fname)
            return
        # frd reading uses formatting from ccx 2.8 or higher so
        # throw an exception if our version is too old
        self.check_ccx_version(timeout=1)
        infile = open(fname, 'r')
        print('Loading nodal results from file: '+fname)
        mode = None
        time = 0.0
        rfstr = ''
        while True:
            line = infile.readline()
            if not line:
                break

            # set the results mode
            if '1PSTEP' in line:
                # we are in a results block
                arr = self.__modearr_nresults(infile)
                line, mode, rfstr, time = arr

            # set mode to none if we hit the end of a resuls block
            if line[:3] == ' -3':
                mode = None
            if not mode:
                continue

            node_data_saver = getattr(self, '_save_node_' + mode)
            node_data_saver(line, rfstr, time)

        infile.close()
        print('The following times have been read:')
        print(self.__steps)
        print('Nodal results from file: %s have been read.' % fname)
        self.set_time(self.__steps[0])

    def __read_dat(self):
        """
        Reads ccx dat results file.
        It has element integration point results.
        """
        fname = self.__problem.fname+'.dat'
        if not os.path.isfile(fname):
            print('Error: %s file not found' % fname)
            return
        infile = open(fname, 'r')
        print('Loading element results from file: '+fname)
        mode = None
        rfstr = ''
        time = 0.0
        while True:
            line = infile.readline()
            if not line:
                break

            # check for stress, we skip down to the line data when
            # we call __modearr_estrsresults
            if 'stress' in line:
                arr = self.__modearr_estrsresults(infile, line)
                line, mode, rfstr, time = arr

            # reset the read type if we hit a blank line
            if line.strip() == '':
                mode = None
            if not mode:
                continue

            # store stress results
            self._save_ele_stress(line, rfstr, time)

        infile.close()

        # loop over all element results, calculating avg element result
        # by averaging integration point vals
        for time in self.__steps:
            for edict in self.__results[time]['element'].values():
                ipoints = edict['ipoints'].values()
                stress_types = ['Sx', 'Sy', 'Sz', 'Sxy', 'Sxz', 'Syz']
                strslist_by_strstype = collections.defaultdict(list)
                # set stress values in max, min, avg locations
                # of non-summary stress components
                for stress_type in stress_types:
                    stress_vals = [ipt[stress_type] for ipt in ipoints]
                    stress_avg = sum(stress_vals)/len(stress_vals)
                    stress_max = max(stress_vals)
                    stress_min =  min(stress_vals)
                    edict['avg'][stress_type] = stress_avg
                    edict['max'][stress_type] = stress_max
                    edict['min'][stress_type] = stress_min
                    strslist_by_strstype[stress_type] = stress_vals
                # for each element, calc Seqv, S1, S2, S3
                # at each integration point
                for ipt in ipoints:
                    stress_vals = [ipt[stress_type] for stress_type
                                   in stress_types]
                    seqv = self.__seqv(stress_vals)
                    [s_1, s_2, s_3] = self.__principals(stress_vals)
                    strslist_by_strstype['Seqv'].append(seqv)
                    strslist_by_strstype['S1'].append(s_1)
                    strslist_by_strstype['S2'].append(s_2)
                    strslist_by_strstype['S3'].append(s_3)
                # now at the element level, store the avg, max, and min
                # of the Seqv and principal stresses we found at
                # integration points
                for stress_type in ['Seqv', 'S1', 'S2', 'S3']:
                    stress_vals = strslist_by_strstype[stress_type]
                    stress_avg = sum(stress_vals)/len(stress_vals)
                    stress_max = max(stress_vals)
                    stress_min =  min(stress_vals)
                    edict['avg'][stress_type] = stress_avg
                    edict['max'][stress_type] = stress_max
                    edict['min'][stress_type] = stress_min


        print('The following times have been read:')
        print(self.__steps)
        print('Element results from file: %s have been read.' % fname)
